# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# This code parses date/times, so please
#
#     pip install python-dateutil
#
# To use this code, make sure you
#
#     import json
#
# and then, to convert JSON from a string, do
#
#     result = auth_event_data_from_dict(json.loads(json_string))

from typing import Optional, Dict, Any, List, TypeVar, Callable, Type, cast
from datetime import datetime
import dateutil.parser


T = TypeVar("T")


def from_dict(f: Callable[[Any], T], x: Any) -> Dict[str, T]:
    assert isinstance(x, dict)
    return { k: f(v) for (k, v) in x.items() }


def from_none(x: Any) -> Any:
    assert x is None
    return x


def from_union(fs, x):
    for f in fs:
        try:
            return f(x)
        except:
            pass
    assert False


def from_datetime(x: Any) -> datetime:
    return dateutil.parser.parse(x)


def from_str(x: Any) -> str:
    assert isinstance(x, str)
    return x


def from_bool(x: Any) -> bool:
    assert isinstance(x, bool)
    return x


def from_list(f: Callable[[Any], T], x: Any) -> List[T]:
    assert isinstance(x, list)
    return [f(y) for y in x]


def to_class(c: Type[T], x: Any) -> dict:
    assert isinstance(x, c)
    return cast(Any, x).to_dict()


class CustomClaims:
    """User's custom claims, typically used to define user roles and propagated
    to an authenticated user's ID token.
    """
    """Unordered map of dynamically typed values."""
    fields: Optional[Dict[str, Optional[Dict[str, Any]]]]

    def __init__(self, fields: Optional[Dict[str, Optional[Dict[str, Any]]]]) -> None:
        self.fields = fields

    @staticmethod
    def from_dict(obj: Any) -> 'CustomClaims':
        assert isinstance(obj, dict)
        fields = from_union([lambda x: from_dict(lambda x: from_union([from_none, lambda x: from_dict(lambda x: x, x)], x), x), from_none], obj.get("fields"))
        return CustomClaims(fields)

    def to_dict(self) -> dict:
        result: dict = {}
        result["fields"] = from_union([lambda x: from_dict(lambda x: from_union([from_none, lambda x: from_dict(lambda x: x, x)], x), x), from_none], self.fields)
        return result


class Metadata:
    """Additional metadata about the user."""
    """The date the user was created."""
    created_at: Optional[datetime]
    """The date the user last signed in."""
    last_signed_in_at: Optional[datetime]

    def __init__(self, created_at: Optional[datetime], last_signed_in_at: Optional[datetime]) -> None:
        self.created_at = created_at
        self.last_signed_in_at = last_signed_in_at

    @staticmethod
    def from_dict(obj: Any) -> 'Metadata':
        assert isinstance(obj, dict)
        created_at = from_union([from_datetime, from_none], obj.get("createdAt"))
        last_signed_in_at = from_union([from_datetime, from_none], obj.get("lastSignedInAt"))
        return Metadata(created_at, last_signed_in_at)

    def to_dict(self) -> dict:
        result: dict = {}
        result["createdAt"] = from_union([lambda x: x.isoformat(), from_none], self.created_at)
        result["lastSignedInAt"] = from_union([lambda x: x.isoformat(), from_none], self.last_signed_in_at)
        return result


class ProviderDatum:
    """User's info at the identity provider"""
    """The display name for the linked provider."""
    display_name: Optional[str]
    """The email for the linked provider."""
    email: Optional[str]
    """The photo URL for the linked provider."""
    photo_url: Optional[str]
    """The linked provider ID (e.g. "google.com" for the Google provider)."""
    provider_id: Optional[str]
    """The user identifier for the linked provider."""
    uid: Optional[str]

    def __init__(self, display_name: Optional[str], email: Optional[str], photo_url: Optional[str], provider_id: Optional[str], uid: Optional[str]) -> None:
        self.display_name = display_name
        self.email = email
        self.photo_url = photo_url
        self.provider_id = provider_id
        self.uid = uid

    @staticmethod
    def from_dict(obj: Any) -> 'ProviderDatum':
        assert isinstance(obj, dict)
        display_name = from_union([from_str, from_none], obj.get("displayName"))
        email = from_union([from_str, from_none], obj.get("email"))
        photo_url = from_union([from_str, from_none], obj.get("photoUrl"))
        provider_id = from_union([from_str, from_none], obj.get("providerId"))
        uid = from_union([from_str, from_none], obj.get("uid"))
        return ProviderDatum(display_name, email, photo_url, provider_id, uid)

    def to_dict(self) -> dict:
        result: dict = {}
        result["displayName"] = from_union([from_str, from_none], self.display_name)
        result["email"] = from_union([from_str, from_none], self.email)
        result["photoUrl"] = from_union([from_str, from_none], self.photo_url)
        result["providerId"] = from_union([from_str, from_none], self.provider_id)
        result["uid"] = from_union([from_str, from_none], self.uid)
        return result


class AuthEventData:
    """The data within all Firebase Auth events."""
    """User's custom claims, typically used to define user roles and propagated
    to an authenticated user's ID token.
    """
    custom_claims: Optional[CustomClaims]
    """Whether the user is disabled."""
    disabled: Optional[bool]
    """The user's display name."""
    display_name: Optional[str]
    """The user's primary email, if set."""
    email: Optional[str]
    """Whether or not the user's primary email is verified."""
    email_verified: Optional[bool]
    """Additional metadata about the user."""
    metadata: Optional[Metadata]
    """The user's phone number."""
    phone_number: Optional[str]
    """The user's photo URL."""
    photo_url: Optional[str]
    """User's info at the providers"""
    provider_data: Optional[List[ProviderDatum]]
    """The user identifier in the Firebase app."""
    uid: Optional[str]

    def __init__(self, custom_claims: Optional[CustomClaims], disabled: Optional[bool], display_name: Optional[str], email: Optional[str], email_verified: Optional[bool], metadata: Optional[Metadata], phone_number: Optional[str], photo_url: Optional[str], provider_data: Optional[List[ProviderDatum]], uid: Optional[str]) -> None:
        self.custom_claims = custom_claims
        self.disabled = disabled
        self.display_name = display_name
        self.email = email
        self.email_verified = email_verified
        self.metadata = metadata
        self.phone_number = phone_number
        self.photo_url = photo_url
        self.provider_data = provider_data
        self.uid = uid

    @staticmethod
    def from_dict(obj: Any) -> 'AuthEventData':
        assert isinstance(obj, dict)
        custom_claims = from_union([CustomClaims.from_dict, from_none], obj.get("customClaims"))
        disabled = from_union([from_bool, from_none], obj.get("disabled"))
        display_name = from_union([from_str, from_none], obj.get("displayName"))
        email = from_union([from_str, from_none], obj.get("email"))
        email_verified = from_union([from_bool, from_none], obj.get("emailVerified"))
        metadata = from_union([Metadata.from_dict, from_none], obj.get("metadata"))
        phone_number = from_union([from_str, from_none], obj.get("phoneNumber"))
        photo_url = from_union([from_str, from_none], obj.get("photoUrl"))
        provider_data = from_union([lambda x: from_list(ProviderDatum.from_dict, x), from_none], obj.get("providerData"))
        uid = from_union([from_str, from_none], obj.get("uid"))
        return AuthEventData(custom_claims, disabled, display_name, email, email_verified, metadata, phone_number, photo_url, provider_data, uid)

    def to_dict(self) -> dict:
        result: dict = {}
        result["customClaims"] = from_union([lambda x: to_class(CustomClaims, x), from_none], self.custom_claims)
        result["disabled"] = from_union([from_bool, from_none], self.disabled)
        result["displayName"] = from_union([from_str, from_none], self.display_name)
        result["email"] = from_union([from_str, from_none], self.email)
        result["emailVerified"] = from_union([from_bool, from_none], self.email_verified)
        result["metadata"] = from_union([lambda x: to_class(Metadata, x), from_none], self.metadata)
        result["phoneNumber"] = from_union([from_str, from_none], self.phone_number)
        result["photoUrl"] = from_union([from_str, from_none], self.photo_url)
        result["providerData"] = from_union([lambda x: from_list(lambda x: to_class(ProviderDatum, x), x), from_none], self.provider_data)
        result["uid"] = from_union([from_str, from_none], self.uid)
        return result


def auth_event_data_from_dict(s: Any) -> AuthEventData:
    return AuthEventData.from_dict(s)


def auth_event_data_to_dict(x: AuthEventData) -> Any:
    return to_class(AuthEventData, x)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# This code parses date/times, so please
#
#     pip install python-dateutil
#
# To use this code, make sure you
#
#     import json
#
# and then, to convert JSON from a string, do
#
#     result = storage_object_data_from_dict(json.loads(json_string))

from typing import Optional, Any, Dict, TypeVar, Callable, Type, cast
from datetime import datetime
import dateutil.parser


T = TypeVar("T")


def from_str(x: Any) -> str:
    assert isinstance(x, str)
    return x


def from_none(x: Any) -> Any:
    assert x is None
    return x


def from_union(fs, x):
    for f in fs:
        try:
            return f(x)
        except:
            pass
    assert False


def from_int(x: Any) -> int:
    assert isinstance(x, int) and not isinstance(x, bool)
    return x


def from_bool(x: Any) -> bool:
    assert isinstance(x, bool)
    return x


def from_dict(f: Callable[[Any], T], x: Any) -> Dict[str, T]:
    assert isinstance(x, dict)
    return { k: f(v) for (k, v) in x.items() }


def from_datetime(x: Any) -> datetime:
    return dateutil.parser.parse(x)


def to_class(c: Type[T], x: Any) -> dict:
    assert isinstance(x, c)
    return cast(Any, x).to_dict()


class CustomerEncryption:
    """Metadata of customer-supplied encryption key, if the object is encrypted by
    such a key.
    """
    """The encryption algorithm."""
    encryption_algorithm: Optional[str]
    """SHA256 hash value of the encryption key."""
    key_sha256: Optional[str]

    def __init__(self, encryption_algorithm: Optional[str], key_sha256: Optional[str]) -> None:
        self.encryption_algorithm = encryption_algorithm
        self.key_sha256 = key_sha256

    @staticmethod
    def from_dict(obj: Any) -> 'CustomerEncryption':
        assert isinstance(obj, dict)
        encryption_algorithm = from_union([from_str, from_none], obj.get("encryptionAlgorithm"))
        key_sha256 = from_union([from_str, from_none], obj.get("keySha256"))
        return CustomerEncryption(encryption_algorithm, key_sha256)

    def to_dict(self) -> dict:
        result: dict = {}
        result["encryptionAlgorithm"] = from_union([from_str, from_none], self.encryption_algorithm)
        result["keySha256"] = from_union([from_str, from_none], self.key_sha256)
        return result


class StorageObjectData:
    """An object within Google Cloud Storage."""
    """The name of the bucket containing this object."""
    bucket: Optional[str]
    """Cache-Control directive for the object data, matching
    [https://tools.ietf.org/html/rfc7234#section-5.2"][RFC 7234 §5.2].
    """
    cache_control: Optional[str]
    """Number of underlying components that make up this object. Components are
    accumulated by compose operations.
    Attempting to set this field will result in an error.
    """
    component_count: Optional[int]
    """Content-Disposition of the object data, matching
    [https://tools.ietf.org/html/rfc6266][RFC 6266].
    """
    content_disposition: Optional[str]
    """Content-Encoding of the object data, matching
    [https://tools.ietf.org/html/rfc7231#section-3.1.2.2][RFC 7231 §3.1.2.2]
    """
    content_encoding: Optional[str]
    """Content-Language of the object data, matching
    [https://tools.ietf.org/html/rfc7231#section-3.1.3.2][RFC 7231 §3.1.3.2].
    """
    content_language: Optional[str]
    """Content-Type of the object data, matching
    [https://tools.ietf.org/html/rfc7231#section-3.1.1.5][RFC 7231 §3.1.1.5].
    If an object is stored without a Content-Type, it is served as
    `application/octet-stream`.
    """
    content_type: Optional[str]
    """CRC32c checksum. For more information about using the CRC32c
    checksum, see
    [https://cloud.google.com/storage/docs/hashes-etags#_JSONAPI][Hashes and
    ETags: Best Practices].
    """
    crc32_c: Optional[str]
    """Metadata of customer-supplied encryption key, if the object is encrypted by
    such a key.
    """
    customer_encryption: Optional[CustomerEncryption]
    """HTTP 1.1 Entity tag for the object. See
    [https://tools.ietf.org/html/rfc7232#section-2.3][RFC 7232 §2.3].
    """
    etag: Optional[str]
    """Whether an object is under event-based hold."""
    event_based_hold: Optional[bool]
    """The content generation of this object. Used for object versioning.
    Attempting to set this field will result in an error.
    """
    generation: Optional[str]
    """The ID of the object, including the bucket name, object name, and
    generation number.
    """
    id: Optional[str]
    """The kind of item this is. For objects, this is always "storage#object"."""
    kind: Optional[str]
    """Cloud KMS Key used to encrypt this object, if the object is encrypted by
    such a key.
    """
    kms_key_name: Optional[str]
    """MD5 hash of the data; encoded using base64 as per
    [https://tools.ietf.org/html/rfc4648#section-4][RFC 4648 §4]. For more
    information about using the MD5 hash, see
    [https://cloud.google.com/storage/docs/hashes-etags#_JSONAPI][Hashes and
    ETags: Best Practices].
    """
    md5_hash: Optional[str]
    """Media download link."""
    media_link: Optional[str]
    """User-provided metadata, in key/value pairs."""
    metadata: Optional[Dict[str, str]]
    """The version of the metadata for this object at this generation. Used for
    preconditions and for detecting changes in metadata. A metageneration
    number is only meaningful in the context of a particular generation of a
    particular object.
    """
    metageneration: Optional[str]
    """The name of the object."""
    name: Optional[str]
    """A server-determined value that specifies the earliest time that the
    object's retention period expires.
    """
    retention_expiration_time: Optional[datetime]
    """The link to this object."""
    self_link: Optional[str]
    """Content-Length of the object data in bytes, matching
    [https://tools.ietf.org/html/rfc7230#section-3.3.2][RFC 7230 §3.3.2].
    """
    size: Optional[str]
    """Storage class of the object."""
    storage_class: Optional[str]
    """Whether an object is under temporary hold."""
    temporary_hold: Optional[bool]
    """The creation time of the object.
    Attempting to set this field will result in an error.
    """
    time_created: Optional[datetime]
    """The deletion time of the object. Will be returned if and only if this
    version of the object has been deleted.
    """
    time_deleted: Optional[datetime]
    """The time at which the object's storage class was last changed."""
    time_storage_class_updated: Optional[datetime]
    """The modification time of the object metadata."""
    updated: Optional[datetime]

    def __init__(self, bucket: Optional[str], cache_control: Optional[str], component_count: Optional[int], content_disposition: Optional[str], content_encoding: Optional[str], content_language: Optional[str], content_type: Optional[str], crc32_c: Optional[str], customer_encryption: Optional[CustomerEncryption], etag: Optional[str], event_based_hold: Optional[bool], generation: Optional[str], id: Optional[str], kind: Optional[str], kms_key_name: Optional[str], md5_hash: Optional[str], media_link: Optional[str], metadata: Optional[Dict[str, str]], metageneration: Optional[str], name: Optional[str], retention_expiration_time: Optional[datetime], self_link: Optional[str], size: Optional[str], storage_class: Optional[str], temporary_hold: Optional[bool], time_created: Optional[datetime], time_deleted: Optional[datetime], time_storage_class_updated: Optional[datetime], updated: Optional[datetime]) -> None:
        self.bucket = bucket
        self.cache_control = cache_control
        self.component_count = component_count
        self.content_disposition = content_disposition
        self.content_encoding = content_encoding
        self.content_language = content_language
        self.content_type = content_type
        self.crc32_c = crc32_c
        self.customer_encryption = customer_encryption
        self.etag = etag
        self.event_based_hold = event_based_hold
        self.generation = generation
        self.id = id
        self.kind = kind
        self.kms_key_name = kms_key_name
        self.md5_hash = md5_hash
        self.media_link = media_link
        self.metadata = metadata
        self.metageneration = metageneration
        self.name = name
        self.retention_expiration_time = retention_expiration_time
        self.self_link = self_link
        self.size = size
        self.storage_class = storage_class
        self.temporary_hold = temporary_hold
        self.time_created = time_created
        self.time_deleted = time_deleted
        self.time_storage_class_updated = time_storage_class_updated
        self.updated = updated

    @staticmethod
    def from_dict(obj: Any) -> 'StorageObjectData':
        assert isinstance(obj, dict)
        bucket = from_union([from_str, from_none], obj.get("bucket"))
        cache_control = from_union([from_str, from_none], obj.get("cacheControl"))
        component_count = from_union([from_int, from_none], obj.get("componentCount"))
        content_disposition = from_union([from_str, from_none], obj.get("contentDisposition"))
        content_encoding = from_union([from_str, from_none], obj.get("contentEncoding"))
        content_language = from_union([from_str, from_none], obj.get("contentLanguage"))
        content_type = from_union([from_str, from_none], obj.get("contentType"))
        crc32_c = from_union([from_str, from_none], obj.get("crc32C"))
        customer_encryption = from_union([CustomerEncryption.from_dict, from_none], obj.get("customerEncryption"))
        etag = from_union([from_str, from_none], obj.get("etag"))
        event_based_hold = from_union([from_bool, from_none], obj.get("eventBasedHold"))
        generation = from_union([from_str, from_none], obj.get("generation"))
        id = from_union([from_str, from_none], obj.get("id"))
        kind = from_union([from_str, from_none], obj.get("kind"))
        kms_key_name = from_union([from_str, from_none], obj.get("kmsKeyName"))
        md5_hash = from_union([from_str, from_none], obj.get("md5Hash"))
        media_link = from_union([from_str, from_none], obj.get("mediaLink"))
        metadata = from_union([lambda x: from_dict(from_str, x), from_none], obj.get("metadata"))
        metageneration = from_union([from_str, from_none], obj.get("metageneration"))
        name = from_union([from_str, from_none], obj.get("name"))
        retention_expiration_time = from_union([from_datetime, from_none], obj.get("retentionExpirationTime"))
        self_link = from_union([from_str, from_none], obj.get("selfLink"))
        size = from_union([from_str, from_none], obj.get("size"))
        storage_class = from_union([from_str, from_none], obj.get("storageClass"))
        temporary_hold = from_union([from_bool, from_none], obj.get("temporaryHold"))
        time_created = from_union([from_datetime, from_none], obj.get("timeCreated"))
        time_deleted = from_union([from_datetime, from_none], obj.get("timeDeleted"))
        time_storage_class_updated = from_union([from_datetime, from_none], obj.get("timeStorageClassUpdated"))
        updated = from_union([from_datetime, from_none], obj.get("updated"))
        return StorageObjectData(bucket, cache_control, component_count, content_disposition, content_encoding, content_language, content_type, crc32_c, customer_encryption, etag, event_based_hold, generation, id, kind, kms_key_name, md5_hash, media_link, metadata, metageneration, name, retention_expiration_time, self_link, size, storage_class, temporary_hold, time_created, time_deleted, time_storage_class_updated, updated)

    def to_dict(self) -> dict:
        result: dict = {}
        result["bucket"] = from_union([from_str, from_none], self.bucket)
        result["cacheControl"] = from_union([from_str, from_none], self.cache_control)
        result["componentCount"] = from_union([from_int, from_none], self.component_count)
        result["contentDisposition"] = from_union([from_str, from_none], self.content_disposition)
        result["contentEncoding"] = from_union([from_str, from_none], self.content_encoding)
        result["contentLanguage"] = from_union([from_str, from_none], self.content_language)
        result["contentType"] = from_union([from_str, from_none], self.content_type)
        result["crc32C"] = from_union([from_str, from_none], self.crc32_c)
        result["customerEncryption"] = from_union([lambda x: to_class(CustomerEncryption, x), from_none], self.customer_encryption)
        result["etag"] = from_union([from_str, from_none], self.etag)
        result["eventBasedHold"] = from_union([from_bool, from_none], self.event_based_hold)
        result["generation"] = from_union([from_str, from_none], self.generation)
        result["id"] = from_union([from_str, from_none], self.id)
        result["kind"] = from_union([from_str, from_none], self.kind)
        result["kmsKeyName"] = from_union([from_str, from_none], self.kms_key_name)
        result["md5Hash"] = from_union([from_str, from_none], self.md5_hash)
        result["mediaLink"] = from_union([from_str, from_none], self.media_link)
        result["metadata"] = from_union([lambda x: from_dict(from_str, x), from_none], self.metadata)
        result["metageneration"] = from_union([from_str, from_none], self.metageneration)
        result["name"] = from_union([from_str, from_none], self.name)
        result["retentionExpirationTime"] = from_union([lambda x: x.isoformat(), from_none], self.retention_expiration_time)
        result["selfLink"] = from_union([from_str, from_none], self.self_link)
        result["size"] = from_union([from_str, from_none], self.size)
        result["storageClass"] = from_union([from_str, from_none], self.storage_class)
        result["temporaryHold"] = from_union([from_bool, from_none], self.temporary_hold)
        result["timeCreated"] = from_union([lambda x: x.isoformat(), from_none], self.time_created)
        result["timeDeleted"] = from_union([lambda x: x.isoformat(), from_none], self.time_deleted)
        result["timeStorageClassUpdated"] = from_union([lambda x: x.isoformat(), from_none], self.time_storage_class_updated)
        result["updated"] = from_union([lambda x: x.isoformat(), from_none], self.updated)
        return result


def storage_object_data_from_dict(s: Any) -> StorageObjectData:
    return StorageObjectData.from_dict(s)


def storage_object_data_to_dict(x: StorageObjectData) -> Any:
    return to_class(StorageObjectData, x)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# This code parses date/times, so please
#
#     pip install python-dateutil
#
# To use this code, make sure you
#
#     import json
#
# and then, to convert JSON from a string, do
#
#     result = log_entry_data_from_dict(json.loads(json_string))

from typing import Optional, Any, Dict, List, Union, TypeVar, Callable, Type, cast
from datetime import datetime
import dateutil.parser


T = TypeVar("T")


def from_bool(x: Any) -> bool:
    assert isinstance(x, bool)
    return x


def from_none(x: Any) -> Any:
    assert x is None
    return x


def from_union(fs, x):
    for f in fs:
        try:
            return f(x)
        except:
            pass
    assert False


def from_str(x: Any) -> str:
    assert isinstance(x, str)
    return x


def from_dict(f: Callable[[Any], T], x: Any) -> Dict[str, T]:
    assert isinstance(x, dict)
    return { k: f(v) for (k, v) in x.items() }


def to_class(c: Type[T], x: Any) -> dict:
    assert isinstance(x, c)
    return cast(Any, x).to_dict()


def from_list(f: Callable[[Any], T], x: Any) -> List[T]:
    assert isinstance(x, list)
    return [f(y) for y in x]


def from_datetime(x: Any) -> datetime:
    return dateutil.parser.parse(x)


def from_int(x: Any) -> int:
    assert isinstance(x, int) and not isinstance(x, bool)
    return x


class Operation:
    """Information about an operation associated with the log entry, if applicable."""
    """True if this is the first log entry in the operation."""
    first: Optional[bool]
    """An arbitrary operation identifier. Log entries with the same
    identifier are assumed to be part of the same operation.
    """
    id: Optional[str]
    """True if this is the last log entry in the operation."""
    last: Optional[bool]
    """An arbitrary producer identifier. The combination of `id` and
    `producer` must be globally unique. Examples for `producer`:
    `"MyDivision.MyBigCompany.com"`, `"github.com/MyProject/MyApplication"`.
    """
    producer: Optional[str]

    def __init__(self, first: Optional[bool], id: Optional[str], last: Optional[bool], producer: Optional[str]) -> None:
        self.first = first
        self.id = id
        self.last = last
        self.producer = producer

    @staticmethod
    def from_dict(obj: Any) -> 'Operation':
        assert isinstance(obj, dict)
        first = from_union([from_bool, from_none], obj.get("first"))
        id = from_union([from_str, from_none], obj.get("id"))
        last = from_union([from_bool, from_none], obj.get("last"))
        producer = from_union([from_str, from_none], obj.get("producer"))
        return Operation(first, id, last, producer)

    def to_dict(self) -> dict:
        result: dict = {}
        result["first"] = from_union([from_bool, from_none], self.first)
        result["id"] = from_union([from_str, from_none], self.id)
        result["last"] = from_union([from_bool, from_none], self.last)
        result["producer"] = from_union([from_str, from_none], self.producer)
        return result


class ServiceMetadata:
    """Metadata about the service that uses the service account."""
    """Unordered map of dynamically typed values."""
    fields: Optional[Dict[str, Optional[Dict[str, Any]]]]

    def __init__(self, fields: Optional[Dict[str, Optional[Dict[str, Any]]]]) -> None:
        self.fields = fields

    @staticmethod
    def from_dict(obj: Any) -> 'ServiceMetadata':
        assert isinstance(obj, dict)
        fields = from_union([lambda x: from_dict(lambda x: from_union([from_none, lambda x: from_dict(lambda x: x, x)], x), x), from_none], obj.get("fields"))
        return ServiceMetadata(fields)

    def to_dict(self) -> dict:
        result: dict = {}
        result["fields"] = from_union([lambda x: from_dict(lambda x: from_union([from_none, lambda x: from_dict(lambda x: x, x)], x), x), from_none], self.fields)
        return result


class FirstPartyPrincipal:
    """First party (Google) identity as the real authority."""
    """The email address of a Google account."""
    principal_email: Optional[str]
    """Metadata about the service that uses the service account."""
    service_metadata: Optional[ServiceMetadata]

    def __init__(self, principal_email: Optional[str], service_metadata: Optional[ServiceMetadata]) -> None:
        self.principal_email = principal_email
        self.service_metadata = service_metadata

    @staticmethod
    def from_dict(obj: Any) -> 'FirstPartyPrincipal':
        assert isinstance(obj, dict)
        principal_email = from_union([from_str, from_none], obj.get("principalEmail"))
        service_metadata = from_union([ServiceMetadata.from_dict, from_none], obj.get("serviceMetadata"))
        return FirstPartyPrincipal(principal_email, service_metadata)

    def to_dict(self) -> dict:
        result: dict = {}
        result["principalEmail"] = from_union([from_str, from_none], self.principal_email)
        result["serviceMetadata"] = from_union([lambda x: to_class(ServiceMetadata, x), from_none], self.service_metadata)
        return result


class ThirdPartyClaims:
    """Metadata about third party identity."""
    """Unordered map of dynamically typed values."""
    fields: Optional[Dict[str, Optional[Dict[str, Any]]]]

    def __init__(self, fields: Optional[Dict[str, Optional[Dict[str, Any]]]]) -> None:
        self.fields = fields

    @staticmethod
    def from_dict(obj: Any) -> 'ThirdPartyClaims':
        assert isinstance(obj, dict)
        fields = from_union([lambda x: from_dict(lambda x: from_union([from_none, lambda x: from_dict(lambda x: x, x)], x), x), from_none], obj.get("fields"))
        return ThirdPartyClaims(fields)

    def to_dict(self) -> dict:
        result: dict = {}
        result["fields"] = from_union([lambda x: from_dict(lambda x: from_union([from_none, lambda x: from_dict(lambda x: x, x)], x), x), from_none], self.fields)
        return result


class ServiceAccountDelegationInfoThirdPartyPrincipal:
    """Third party identity as the real authority."""
    """Metadata about third party identity."""
    third_party_claims: Optional[ThirdPartyClaims]

    def __init__(self, third_party_claims: Optional[ThirdPartyClaims]) -> None:
        self.third_party_claims = third_party_claims

    @staticmethod
    def from_dict(obj: Any) -> 'ServiceAccountDelegationInfoThirdPartyPrincipal':
        assert isinstance(obj, dict)
        third_party_claims = from_union([ThirdPartyClaims.from_dict, from_none], obj.get("thirdPartyClaims"))
        return ServiceAccountDelegationInfoThirdPartyPrincipal(third_party_claims)

    def to_dict(self) -> dict:
        result: dict = {}
        result["thirdPartyClaims"] = from_union([lambda x: to_class(ThirdPartyClaims, x), from_none], self.third_party_claims)
        return result


class ServiceAccountDelegationInfo:
    """Identity delegation history of an authenticated service account."""
    """First party (Google) identity as the real authority."""
    first_party_principal: Optional[FirstPartyPrincipal]
    """Third party identity as the real authority."""
    third_party_principal: Optional[ServiceAccountDelegationInfoThirdPartyPrincipal]

    def __init__(self, first_party_principal: Optional[FirstPartyPrincipal], third_party_principal: Optional[ServiceAccountDelegationInfoThirdPartyPrincipal]) -> None:
        self.first_party_principal = first_party_principal
        self.third_party_principal = third_party_principal

    @staticmethod
    def from_dict(obj: Any) -> 'ServiceAccountDelegationInfo':
        assert isinstance(obj, dict)
        first_party_principal = from_union([FirstPartyPrincipal.from_dict, from_none], obj.get("firstPartyPrincipal"))
        third_party_principal = from_union([ServiceAccountDelegationInfoThirdPartyPrincipal.from_dict, from_none], obj.get("thirdPartyPrincipal"))
        return ServiceAccountDelegationInfo(first_party_principal, third_party_principal)

    def to_dict(self) -> dict:
        result: dict = {}
        result["firstPartyPrincipal"] = from_union([lambda x: to_class(FirstPartyPrincipal, x), from_none], self.first_party_principal)
        result["thirdPartyPrincipal"] = from_union([lambda x: to_class(ServiceAccountDelegationInfoThirdPartyPrincipal, x), from_none], self.third_party_principal)
        return result


class AuthenticationInfoThirdPartyPrincipal:
    """The third party identification (if any) of the authenticated user making
    the request.
    When the JSON object represented here has a proto equivalent, the proto
    name will be indicated in the `@type` property.
    """
    """Unordered map of dynamically typed values."""
    fields: Optional[Dict[str, Optional[Dict[str, Any]]]]

    def __init__(self, fields: Optional[Dict[str, Optional[Dict[str, Any]]]]) -> None:
        self.fields = fields

    @staticmethod
    def from_dict(obj: Any) -> 'AuthenticationInfoThirdPartyPrincipal':
        assert isinstance(obj, dict)
        fields = from_union([lambda x: from_dict(lambda x: from_union([from_none, lambda x: from_dict(lambda x: x, x)], x), x), from_none], obj.get("fields"))
        return AuthenticationInfoThirdPartyPrincipal(fields)

    def to_dict(self) -> dict:
        result: dict = {}
        result["fields"] = from_union([lambda x: from_dict(lambda x: from_union([from_none, lambda x: from_dict(lambda x: x, x)], x), x), from_none], self.fields)
        return result


class AuthenticationInfo:
    """Authentication information."""
    """The authority selector specified by the requestor, if any.
    It is not guaranteed that the principal was allowed to use this authority.
    """
    authority_selector: Optional[str]
    """The email address of the authenticated user (or service account on behalf
    of third party principal) making the request. For privacy reasons, the
    principal email address is redacted for all read-only operations that fail
    with a "permission denied" error.
    """
    principal_email: Optional[str]
    """String representation of identity of requesting party.
    Populated for both first and third party identities.
    """
    principal_subject: Optional[str]
    """Identity delegation history of an authenticated service account that makes
    the request. It contains information on the real authorities that try to
    access GCP resources by delegating on a service account. When multiple
    authorities present, they are guaranteed to be sorted based on the original
    ordering of the identity delegation events.
    """
    service_account_delegation_info: Optional[List[ServiceAccountDelegationInfo]]
    """The name of the service account key used to create or exchange
    credentials for authenticating the service account making the request.
    This is a scheme-less URI full resource name. For example:
    
    "//iam.googleapis.com/projects/{PROJECT_ID}/serviceAccounts/{ACCOUNT}/keys/{key}"
    """
    service_account_key_name: Optional[str]
    """The third party identification (if any) of the authenticated user making
    the request.
    When the JSON object represented here has a proto equivalent, the proto
    name will be indicated in the `@type` property.
    """
    third_party_principal: Optional[AuthenticationInfoThirdPartyPrincipal]

    def __init__(self, authority_selector: Optional[str], principal_email: Optional[str], principal_subject: Optional[str], service_account_delegation_info: Optional[List[ServiceAccountDelegationInfo]], service_account_key_name: Optional[str], third_party_principal: Optional[AuthenticationInfoThirdPartyPrincipal]) -> None:
        self.authority_selector = authority_selector
        self.principal_email = principal_email
        self.principal_subject = principal_subject
        self.service_account_delegation_info = service_account_delegation_info
        self.service_account_key_name = service_account_key_name
        self.third_party_principal = third_party_principal

    @staticmethod
    def from_dict(obj: Any) -> 'AuthenticationInfo':
        assert isinstance(obj, dict)
        authority_selector = from_union([from_str, from_none], obj.get("authoritySelector"))
        principal_email = from_union([from_str, from_none], obj.get("principalEmail"))
        principal_subject = from_union([from_str, from_none], obj.get("principalSubject"))
        service_account_delegation_info = from_union([lambda x: from_list(ServiceAccountDelegationInfo.from_dict, x), from_none], obj.get("serviceAccountDelegationInfo"))
        service_account_key_name = from_union([from_str, from_none], obj.get("serviceAccountKeyName"))
        third_party_principal = from_union([AuthenticationInfoThirdPartyPrincipal.from_dict, from_none], obj.get("thirdPartyPrincipal"))
        return AuthenticationInfo(authority_selector, principal_email, principal_subject, service_account_delegation_info, service_account_key_name, third_party_principal)

    def to_dict(self) -> dict:
        result: dict = {}
        result["authoritySelector"] = from_union([from_str, from_none], self.authority_selector)
        result["principalEmail"] = from_union([from_str, from_none], self.principal_email)
        result["principalSubject"] = from_union([from_str, from_none], self.principal_subject)
        result["serviceAccountDelegationInfo"] = from_union([lambda x: from_list(lambda x: to_class(ServiceAccountDelegationInfo, x), x), from_none], self.service_account_delegation_info)
        result["serviceAccountKeyName"] = from_union([from_str, from_none], self.service_account_key_name)
        result["thirdPartyPrincipal"] = from_union([lambda x: to_class(AuthenticationInfoThirdPartyPrincipal, x), from_none], self.third_party_principal)
        return result


class ResourceAttributes:
    """Resource attributes used in IAM condition evaluation. This field contains
    resource attributes like resource type and resource name.
    
    To get the whole view of the attributes used in IAM
    condition evaluation, the user must also look into
    `AuditLogData.request_metadata.request_attributes`.
    """
    """The labels or tags on the resource, such as AWS resource tags and
    Kubernetes resource labels.
    """
    labels: Optional[Dict[str, str]]
    """The stable identifier (name) of a resource on the `service`. A resource
    can be logically identified as "//{resource.service}/{resource.name}".
    The differences between a resource name and a URI are:
    
    *   Resource name is a logical identifier, independent of network
    protocol and API version. For example,
    `//pubsub.googleapis.com/projects/123/topics/news-feed`.
    *   URI often includes protocol and version information, so it can
    be used directly by applications. For example,
    `https://pubsub.googleapis.com/v1/projects/123/topics/news-feed`.
    
    See https://cloud.google.com/apis/design/resource_names for details.
    """
    name: Optional[str]
    """The name of the service that this resource belongs to, such as
    `pubsub.googleapis.com`. The service may be different from the DNS
    hostname that actually serves the request.
    """
    service: Optional[str]
    """The type of the resource. The syntax is platform-specific because
    different platforms define their resources differently.
    
    For Google APIs, the type format must be "{service}/{kind}".
    """
    type: Optional[str]

    def __init__(self, labels: Optional[Dict[str, str]], name: Optional[str], service: Optional[str], type: Optional[str]) -> None:
        self.labels = labels
        self.name = name
        self.service = service
        self.type = type

    @staticmethod
    def from_dict(obj: Any) -> 'ResourceAttributes':
        assert isinstance(obj, dict)
        labels = from_union([lambda x: from_dict(from_str, x), from_none], obj.get("labels"))
        name = from_union([from_str, from_none], obj.get("name"))
        service = from_union([from_str, from_none], obj.get("service"))
        type = from_union([from_str, from_none], obj.get("type"))
        return ResourceAttributes(labels, name, service, type)

    def to_dict(self) -> dict:
        result: dict = {}
        result["labels"] = from_union([lambda x: from_dict(from_str, x), from_none], self.labels)
        result["name"] = from_union([from_str, from_none], self.name)
        result["service"] = from_union([from_str, from_none], self.service)
        result["type"] = from_union([from_str, from_none], self.type)
        return result


class AuthorizationInfo:
    """Authorization information for the operation."""
    """Whether or not authorization for `resource` and `permission`
    was granted.
    """
    granted: Optional[bool]
    """The required IAM permission."""
    permission: Optional[str]
    """The resource being accessed, as a REST-style string. For example:
    
    bigquery.googleapis.com/projects/PROJECTID/datasets/DATASETID
    """
    resource: Optional[str]
    """Resource attributes used in IAM condition evaluation. This field contains
    resource attributes like resource type and resource name.
    
    To get the whole view of the attributes used in IAM
    condition evaluation, the user must also look into
    `AuditLogData.request_metadata.request_attributes`.
    """
    resource_attributes: Optional[ResourceAttributes]

    def __init__(self, granted: Optional[bool], permission: Optional[str], resource: Optional[str], resource_attributes: Optional[ResourceAttributes]) -> None:
        self.granted = granted
        self.permission = permission
        self.resource = resource
        self.resource_attributes = resource_attributes

    @staticmethod
    def from_dict(obj: Any) -> 'AuthorizationInfo':
        assert isinstance(obj, dict)
        granted = from_union([from_bool, from_none], obj.get("granted"))
        permission = from_union([from_str, from_none], obj.get("permission"))
        resource = from_union([from_str, from_none], obj.get("resource"))
        resource_attributes = from_union([ResourceAttributes.from_dict, from_none], obj.get("resourceAttributes"))
        return AuthorizationInfo(granted, permission, resource, resource_attributes)

    def to_dict(self) -> dict:
        result: dict = {}
        result["granted"] = from_union([from_bool, from_none], self.granted)
        result["permission"] = from_union([from_str, from_none], self.permission)
        result["resource"] = from_union([from_str, from_none], self.resource)
        result["resourceAttributes"] = from_union([lambda x: to_class(ResourceAttributes, x), from_none], self.resource_attributes)
        return result


class Metadata:
    """Other service-specific data about the request, response, and other
    information associated with the current audited event.
    """
    """Unordered map of dynamically typed values."""
    fields: Optional[Dict[str, Optional[Dict[str, Any]]]]

    def __init__(self, fields: Optional[Dict[str, Optional[Dict[str, Any]]]]) -> None:
        self.fields = fields

    @staticmethod
    def from_dict(obj: Any) -> 'Metadata':
        assert isinstance(obj, dict)
        fields = from_union([lambda x: from_dict(lambda x: from_union([from_none, lambda x: from_dict(lambda x: x, x)], x), x), from_none], obj.get("fields"))
        return Metadata(fields)

    def to_dict(self) -> dict:
        result: dict = {}
        result["fields"] = from_union([lambda x: from_dict(lambda x: from_union([from_none, lambda x: from_dict(lambda x: x, x)], x), x), from_none], self.fields)
        return result


class Request:
    """The operation request. This may not include all request parameters,
    such as those that are too large, privacy-sensitive, or duplicated
    elsewhere in the log record.
    It should never include user-generated data, such as file contents.
    When the JSON object represented here has a proto equivalent, the proto
    name will be indicated in the `@type` property.
    """
    """Unordered map of dynamically typed values."""
    fields: Optional[Dict[str, Optional[Dict[str, Any]]]]

    def __init__(self, fields: Optional[Dict[str, Optional[Dict[str, Any]]]]) -> None:
        self.fields = fields

    @staticmethod
    def from_dict(obj: Any) -> 'Request':
        assert isinstance(obj, dict)
        fields = from_union([lambda x: from_dict(lambda x: from_union([from_none, lambda x: from_dict(lambda x: x, x)], x), x), from_none], obj.get("fields"))
        return Request(fields)

    def to_dict(self) -> dict:
        result: dict = {}
        result["fields"] = from_union([lambda x: from_dict(lambda x: from_union([from_none, lambda x: from_dict(lambda x: x, x)], x), x), from_none], self.fields)
        return result


class DestinationAttributes:
    """The destination of a network activity, such as accepting a TCP connection.
    In a multi hop network activity, the destination represents the receiver of
    the last hop. Only two fields are used in this message, Peer.port and
    Peer.ip. These fields are optionally populated by those services utilizing
    the IAM condition feature.
    """
    """The IP address of the peer."""
    ip: Optional[str]
    """The labels associated with the peer."""
    labels: Optional[Dict[str, str]]
    """The network port of the peer."""
    port: Optional[str]
    """The identity of this peer. Similar to `Request.auth.principal`, but
    relative to the peer instead of the request. For example, the
    idenity associated with a load balancer that forwared the request.
    """
    principal: Optional[str]
    """The CLDR country/region code associated with the above IP address.
    If the IP address is private, the `region_code` should reflect the
    physical location where this peer is running.
    """
    region_code: Optional[str]

    def __init__(self, ip: Optional[str], labels: Optional[Dict[str, str]], port: Optional[str], principal: Optional[str], region_code: Optional[str]) -> None:
        self.ip = ip
        self.labels = labels
        self.port = port
        self.principal = principal
        self.region_code = region_code

    @staticmethod
    def from_dict(obj: Any) -> 'DestinationAttributes':
        assert isinstance(obj, dict)
        ip = from_union([from_str, from_none], obj.get("ip"))
        labels = from_union([lambda x: from_dict(from_str, x), from_none], obj.get("labels"))
        port = from_union([from_str, from_none], obj.get("port"))
        principal = from_union([from_str, from_none], obj.get("principal"))
        region_code = from_union([from_str, from_none], obj.get("regionCode"))
        return DestinationAttributes(ip, labels, port, principal, region_code)

    def to_dict(self) -> dict:
        result: dict = {}
        result["ip"] = from_union([from_str, from_none], self.ip)
        result["labels"] = from_union([lambda x: from_dict(from_str, x), from_none], self.labels)
        result["port"] = from_union([from_str, from_none], self.port)
        result["principal"] = from_union([from_str, from_none], self.principal)
        result["regionCode"] = from_union([from_str, from_none], self.region_code)
        return result


class Claims:
    """Structured claims presented with the credential. JWTs include
    `{key: value}` pairs for standard and private claims. The following
    is a subset of the standard required and optional claims that would
    typically be presented for a Google-based JWT:
    
    {'iss': 'accounts.google.com',
    'sub': '113289723416554971153',
    'aud': ['123456789012', 'pubsub.googleapis.com'],
    'azp': '123456789012.apps.googleusercontent.com',
    'email': 'jsmith@example.com',
    'iat': 1353601026,
    'exp': 1353604926}
    
    SAML assertions are similarly specified, but with an identity provider
    dependent structure.
    """
    """Unordered map of dynamically typed values."""
    fields: Optional[Dict[str, Optional[Dict[str, Any]]]]

    def __init__(self, fields: Optional[Dict[str, Optional[Dict[str, Any]]]]) -> None:
        self.fields = fields

    @staticmethod
    def from_dict(obj: Any) -> 'Claims':
        assert isinstance(obj, dict)
        fields = from_union([lambda x: from_dict(lambda x: from_union([from_none, lambda x: from_dict(lambda x: x, x)], x), x), from_none], obj.get("fields"))
        return Claims(fields)

    def to_dict(self) -> dict:
        result: dict = {}
        result["fields"] = from_union([lambda x: from_dict(lambda x: from_union([from_none, lambda x: from_dict(lambda x: x, x)], x), x), from_none], self.fields)
        return result


class Auth:
    """The request authentication. May be absent for unauthenticated requests.
    Derived from the HTTP request `Authorization` header or equivalent.
    """
    """A list of access level resource names that allow resources to be
    accessed by authenticated requester. It is part of Secure GCP processing
    for the incoming request. An access level string has the format:
    "//{api_service_name}/accessPolicies/{policy_id}/accessLevels/{short_name}"
    
    Example:
    "//accesscontextmanager.googleapis.com/accessPolicies/MY_POLICY_ID/accessLevels/MY_LEVEL"
    """
    access_levels: Optional[List[str]]
    """The intended audience(s) for this authentication information. Reflects
    the audience (`aud`) claim within a JWT. The audience
    value(s) depends on the `issuer`, but typically include one or more of
    the following pieces of information:
    
    *  The services intended to receive the credential such as
    ["pubsub.googleapis.com", "storage.googleapis.com"]
    *  A set of service-based scopes. For example,
    ["https://www.googleapis.com/auth/cloud-platform"]
    *  The client id of an app, such as the Firebase project id for JWTs
    from Firebase Auth.
    
    Consult the documentation for the credential issuer to determine the
    information provided.
    """
    audiences: Optional[List[str]]
    """Structured claims presented with the credential. JWTs include
    `{key: value}` pairs for standard and private claims. The following
    is a subset of the standard required and optional claims that would
    typically be presented for a Google-based JWT:
    
    {'iss': 'accounts.google.com',
    'sub': '113289723416554971153',
    'aud': ['123456789012', 'pubsub.googleapis.com'],
    'azp': '123456789012.apps.googleusercontent.com',
    'email': 'jsmith@example.com',
    'iat': 1353601026,
    'exp': 1353604926}
    
    SAML assertions are similarly specified, but with an identity provider
    dependent structure.
    """
    claims: Optional[Claims]
    """The authorized presenter of the credential. Reflects the optional
    Authorized Presenter (`azp`) claim within a JWT or the
    OAuth client id. For example, a Google Cloud Platform client id looks
    as follows: "123456789012.apps.googleusercontent.com".
    """
    presenter: Optional[str]
    """The authenticated principal. Reflects the issuer (`iss`) and subject
    (`sub`) claims within a JWT. The issuer and subject should be `/`
    delimited, with `/` percent-encoded within the subject fragment. For
    Google accounts, the principal format is:
    "https://accounts.google.com/{id}"
    """
    principal: Optional[str]

    def __init__(self, access_levels: Optional[List[str]], audiences: Optional[List[str]], claims: Optional[Claims], presenter: Optional[str], principal: Optional[str]) -> None:
        self.access_levels = access_levels
        self.audiences = audiences
        self.claims = claims
        self.presenter = presenter
        self.principal = principal

    @staticmethod
    def from_dict(obj: Any) -> 'Auth':
        assert isinstance(obj, dict)
        access_levels = from_union([lambda x: from_list(from_str, x), from_none], obj.get("accessLevels"))
        audiences = from_union([lambda x: from_list(from_str, x), from_none], obj.get("audiences"))
        claims = from_union([Claims.from_dict, from_none], obj.get("claims"))
        presenter = from_union([from_str, from_none], obj.get("presenter"))
        principal = from_union([from_str, from_none], obj.get("principal"))
        return Auth(access_levels, audiences, claims, presenter, principal)

    def to_dict(self) -> dict:
        result: dict = {}
        result["accessLevels"] = from_union([lambda x: from_list(from_str, x), from_none], self.access_levels)
        result["audiences"] = from_union([lambda x: from_list(from_str, x), from_none], self.audiences)
        result["claims"] = from_union([lambda x: to_class(Claims, x), from_none], self.claims)
        result["presenter"] = from_union([from_str, from_none], self.presenter)
        result["principal"] = from_union([from_str, from_none], self.principal)
        return result


class RequestAttributes:
    """Request attributes used in IAM condition evaluation. This field contains
    request attributes like request time and access levels associated with
    the request.
    
    
    To get the whole view of the attributes used in IAM
    condition evaluation, the user must also look into
    `AuditLog.authentication_info.resource_attributes`.
    """
    """The request authentication. May be absent for unauthenticated requests.
    Derived from the HTTP request `Authorization` header or equivalent.
    """
    auth: Optional[Auth]
    """The HTTP request headers. If multiple headers share the same key, they
    must be merged according to the HTTP spec. All header keys must be
    lowercased, because HTTP header keys are case-insensitive.
    """
    headers: Optional[Dict[str, str]]
    """The HTTP request `Host` header value."""
    host: Optional[str]
    """The unique ID for a request, which can be propagated to downstream
    systems. The ID should have low probability of collision
    within a single day for a specific service.
    """
    id: Optional[str]
    """The HTTP request method, such as `GET`, `POST`."""
    method: Optional[str]
    """The HTTP URL path."""
    path: Optional[str]
    """The network protocol used with the request, such as "http/1.1",
    "spdy/3", "h2", "h2c", "webrtc", "tcp", "udp", "quic". See
    
    https://www.iana.org/assignments/tls-extensiontype-values/tls-extensiontype-values.xhtml#alpn-protocol-ids
    for details.
    """
    protocol: Optional[str]
    """The HTTP URL query in the format of `name1=value1&name2=value2`, as it
    appears in the first line of the HTTP request. No decoding is performed.
    """
    query: Optional[str]
    """A special parameter for request reason. It is used by security systems
    to associate auditing information with a request.
    """
    reason: Optional[str]
    """The HTTP URL scheme, such as `http` and `https`."""
    scheme: Optional[str]
    """The HTTP request size in bytes. If unknown, it must be -1."""
    size: Optional[str]
    """The timestamp when the `destination` service receives the first byte of
    the request.
    """
    time: Optional[datetime]

    def __init__(self, auth: Optional[Auth], headers: Optional[Dict[str, str]], host: Optional[str], id: Optional[str], method: Optional[str], path: Optional[str], protocol: Optional[str], query: Optional[str], reason: Optional[str], scheme: Optional[str], size: Optional[str], time: Optional[datetime]) -> None:
        self.auth = auth
        self.headers = headers
        self.host = host
        self.id = id
        self.method = method
        self.path = path
        self.protocol = protocol
        self.query = query
        self.reason = reason
        self.scheme = scheme
        self.size = size
        self.time = time

    @staticmethod
    def from_dict(obj: Any) -> 'RequestAttributes':
        assert isinstance(obj, dict)
        auth = from_union([Auth.from_dict, from_none], obj.get("auth"))
        headers = from_union([lambda x: from_dict(from_str, x), from_none], obj.get("headers"))
        host = from_union([from_str, from_none], obj.get("host"))
        id = from_union([from_str, from_none], obj.get("id"))
        method = from_union([from_str, from_none], obj.get("method"))
        path = from_union([from_str, from_none], obj.get("path"))
        protocol = from_union([from_str, from_none], obj.get("protocol"))
        query = from_union([from_str, from_none], obj.get("query"))
        reason = from_union([from_str, from_none], obj.get("reason"))
        scheme = from_union([from_str, from_none], obj.get("scheme"))
        size = from_union([from_str, from_none], obj.get("size"))
        time = from_union([from_datetime, from_none], obj.get("time"))
        return RequestAttributes(auth, headers, host, id, method, path, protocol, query, reason, scheme, size, time)

    def to_dict(self) -> dict:
        result: dict = {}
        result["auth"] = from_union([lambda x: to_class(Auth, x), from_none], self.auth)
        result["headers"] = from_union([lambda x: from_dict(from_str, x), from_none], self.headers)
        result["host"] = from_union([from_str, from_none], self.host)
        result["id"] = from_union([from_str, from_none], self.id)
        result["method"] = from_union([from_str, from_none], self.method)
        result["path"] = from_union([from_str, from_none], self.path)
        result["protocol"] = from_union([from_str, from_none], self.protocol)
        result["query"] = from_union([from_str, from_none], self.query)
        result["reason"] = from_union([from_str, from_none], self.reason)
        result["scheme"] = from_union([from_str, from_none], self.scheme)
        result["size"] = from_union([from_str, from_none], self.size)
        result["time"] = from_union([lambda x: x.isoformat(), from_none], self.time)
        return result


class RequestMetadata:
    """Metadata about the operation."""
    """The IP address of the caller.
    For caller from internet, this will be public IPv4 or IPv6 address.
    For caller from a Compute Engine VM with external IP address, this
    will be the VM's external IP address. For caller from a Compute
    Engine VM without external IP address, if the VM is in the same
    organization (or project) as the accessed resource, `caller_ip` will
    be the VM's internal IPv4 address, otherwise the `caller_ip` will be
    redacted to "gce-internal-ip".
    See https://cloud.google.com/compute/docs/vpc/ for more information.
    """
    caller_ip: Optional[str]
    """The network of the caller.
    Set only if the network host project is part of the same GCP organization
    (or project) as the accessed resource.
    See https://cloud.google.com/compute/docs/vpc/ for more information.
    This is a scheme-less URI full resource name. For example:
    
    "//compute.googleapis.com/projects/PROJECT_ID/global/networks/NETWORK_ID"
    """
    caller_network: Optional[str]
    """The user agent of the caller.
    This information is not authenticated and should be treated accordingly.
    For example:
    
    +   `google-api-python-client/1.4.0`:
    The request was made by the Google API client for Python.
    +   `Cloud SDK Command Line Tool apitools-client/1.0 gcloud/0.9.62`:
    The request was made by the Google Cloud SDK CLI (gcloud).
    +   `AppEngine-Google; (+http://code.google.com/appengine; appid:
    s~my-project`:
    The request was made from the `my-project` App Engine app.
    """
    caller_supplied_user_agent: Optional[str]
    """The destination of a network activity, such as accepting a TCP connection.
    In a multi hop network activity, the destination represents the receiver of
    the last hop. Only two fields are used in this message, Peer.port and
    Peer.ip. These fields are optionally populated by those services utilizing
    the IAM condition feature.
    """
    destination_attributes: Optional[DestinationAttributes]
    """Request attributes used in IAM condition evaluation. This field contains
    request attributes like request time and access levels associated with
    the request.
    
    
    To get the whole view of the attributes used in IAM
    condition evaluation, the user must also look into
    `AuditLog.authentication_info.resource_attributes`.
    """
    request_attributes: Optional[RequestAttributes]

    def __init__(self, caller_ip: Optional[str], caller_network: Optional[str], caller_supplied_user_agent: Optional[str], destination_attributes: Optional[DestinationAttributes], request_attributes: Optional[RequestAttributes]) -> None:
        self.caller_ip = caller_ip
        self.caller_network = caller_network
        self.caller_supplied_user_agent = caller_supplied_user_agent
        self.destination_attributes = destination_attributes
        self.request_attributes = request_attributes

    @staticmethod
    def from_dict(obj: Any) -> 'RequestMetadata':
        assert isinstance(obj, dict)
        caller_ip = from_union([from_str, from_none], obj.get("callerIp"))
        caller_network = from_union([from_str, from_none], obj.get("callerNetwork"))
        caller_supplied_user_agent = from_union([from_str, from_none], obj.get("callerSuppliedUserAgent"))
        destination_attributes = from_union([DestinationAttributes.from_dict, from_none], obj.get("destinationAttributes"))
        request_attributes = from_union([RequestAttributes.from_dict, from_none], obj.get("requestAttributes"))
        return RequestMetadata(caller_ip, caller_network, caller_supplied_user_agent, destination_attributes, request_attributes)

    def to_dict(self) -> dict:
        result: dict = {}
        result["callerIp"] = from_union([from_str, from_none], self.caller_ip)
        result["callerNetwork"] = from_union([from_str, from_none], self.caller_network)
        result["callerSuppliedUserAgent"] = from_union([from_str, from_none], self.caller_supplied_user_agent)
        result["destinationAttributes"] = from_union([lambda x: to_class(DestinationAttributes, x), from_none], self.destination_attributes)
        result["requestAttributes"] = from_union([lambda x: to_class(RequestAttributes, x), from_none], self.request_attributes)
        return result


class ResourceLocation:
    """The resource location information."""
    """The locations of a resource after the execution of the operation.
    Requests to create or delete a location based resource must populate
    the 'current_locations' field and not the 'original_locations' field.
    For example:
    
    "europe-west1-a"
    "us-east1"
    "nam3"
    """
    current_locations: Optional[List[str]]
    """The locations of a resource prior to the execution of the operation.
    Requests that mutate the resource's location must populate both the
    'original_locations' as well as the 'current_locations' fields.
    For example:
    
    "europe-west1-a"
    "us-east1"
    "nam3"
    """
    original_locations: Optional[List[str]]

    def __init__(self, current_locations: Optional[List[str]], original_locations: Optional[List[str]]) -> None:
        self.current_locations = current_locations
        self.original_locations = original_locations

    @staticmethod
    def from_dict(obj: Any) -> 'ResourceLocation':
        assert isinstance(obj, dict)
        current_locations = from_union([lambda x: from_list(from_str, x), from_none], obj.get("currentLocations"))
        original_locations = from_union([lambda x: from_list(from_str, x), from_none], obj.get("originalLocations"))
        return ResourceLocation(current_locations, original_locations)

    def to_dict(self) -> dict:
        result: dict = {}
        result["currentLocations"] = from_union([lambda x: from_list(from_str, x), from_none], self.current_locations)
        result["originalLocations"] = from_union([lambda x: from_list(from_str, x), from_none], self.original_locations)
        return result


class ResourceOriginalState:
    """The resource's original state before mutation. Present only for
    operations which have successfully modified the targeted resource(s).
    In general, this field should contain all changed fields, except those
    that are already been included in `request`, `response`, `metadata` or
    `service_data` fields.
    When the JSON object represented here has a proto equivalent,
    the proto name will be indicated in the `@type` property.
    """
    """Unordered map of dynamically typed values."""
    fields: Optional[Dict[str, Optional[Dict[str, Any]]]]

    def __init__(self, fields: Optional[Dict[str, Optional[Dict[str, Any]]]]) -> None:
        self.fields = fields

    @staticmethod
    def from_dict(obj: Any) -> 'ResourceOriginalState':
        assert isinstance(obj, dict)
        fields = from_union([lambda x: from_dict(lambda x: from_union([from_none, lambda x: from_dict(lambda x: x, x)], x), x), from_none], obj.get("fields"))
        return ResourceOriginalState(fields)

    def to_dict(self) -> dict:
        result: dict = {}
        result["fields"] = from_union([lambda x: from_dict(lambda x: from_union([from_none, lambda x: from_dict(lambda x: x, x)], x), x), from_none], self.fields)
        return result


class Response:
    """The operation response. This may not include all response elements,
    such as those that are too large, privacy-sensitive, or duplicated
    elsewhere in the log record.
    It should never include user-generated data, such as file contents.
    When the JSON object represented here has a proto equivalent, the proto
    name will be indicated in the `@type` property.
    """
    """Unordered map of dynamically typed values."""
    fields: Optional[Dict[str, Optional[Dict[str, Any]]]]

    def __init__(self, fields: Optional[Dict[str, Optional[Dict[str, Any]]]]) -> None:
        self.fields = fields

    @staticmethod
    def from_dict(obj: Any) -> 'Response':
        assert isinstance(obj, dict)
        fields = from_union([lambda x: from_dict(lambda x: from_union([from_none, lambda x: from_dict(lambda x: x, x)], x), x), from_none], obj.get("fields"))
        return Response(fields)

    def to_dict(self) -> dict:
        result: dict = {}
        result["fields"] = from_union([lambda x: from_dict(lambda x: from_union([from_none, lambda x: from_dict(lambda x: x, x)], x), x), from_none], self.fields)
        return result


class ServiceData:
    """Deprecated, use `metadata` field instead.
    Other service-specific data about the request, response, and other
    activities.
    When the JSON object represented here has a proto equivalent, the proto
    name will be indicated in the `@type` property.
    """
    """Unordered map of dynamically typed values."""
    fields: Optional[Dict[str, Optional[Dict[str, Any]]]]

    def __init__(self, fields: Optional[Dict[str, Optional[Dict[str, Any]]]]) -> None:
        self.fields = fields

    @staticmethod
    def from_dict(obj: Any) -> 'ServiceData':
        assert isinstance(obj, dict)
        fields = from_union([lambda x: from_dict(lambda x: from_union([from_none, lambda x: from_dict(lambda x: x, x)], x), x), from_none], obj.get("fields"))
        return ServiceData(fields)

    def to_dict(self) -> dict:
        result: dict = {}
        result["fields"] = from_union([lambda x: from_dict(lambda x: from_union([from_none, lambda x: from_dict(lambda x: x, x)], x), x), from_none], self.fields)
        return result


class Detail:
    """`Any` contains an arbitrary serialized protocol buffer message along with a
    URL that describes the type of the serialized message.
    
    Protobuf library provides support to pack/unpack Any values in the form
    of utility functions or additional generated methods of the Any type.
    
    Example 1: Pack and unpack a message in C++.
    
    Foo foo = ...;
    Any any;
    any.PackFrom(foo);
    ...
    if (any.UnpackTo(&foo)) {
    ...
    }
    
    Example 2: Pack and unpack a message in Java.
    
    Foo foo = ...;
    Any any = Any.pack(foo);
    ...
    if (any.is(Foo.class)) {
    foo = any.unpack(Foo.class);
    }
    
    Example 3: Pack and unpack a message in Python.
    
    foo = Foo(...)
    any = Any()
    any.Pack(foo)
    ...
    if any.Is(Foo.DESCRIPTOR):
    any.Unpack(foo)
    ...
    
    Example 4: Pack and unpack a message in Go
    
    foo := &pb.Foo{...}
    any, err := ptypes.MarshalAny(foo)
    ...
    foo := &pb.Foo{}
    if err := ptypes.UnmarshalAny(any, foo); err != nil {
    ...
    }
    
    The pack methods provided by protobuf library will by default use
    'type.googleapis.com/full.type.name' as the type URL and the unpack
    methods only use the fully qualified type name after the last '/'
    in the type URL, for example "foo.bar.com/x/y.z" will yield type
    name "y.z".
    
    
    JSON
    ====
    The JSON representation of an `Any` value uses the regular
    representation of the deserialized, embedded message, with an
    additional field `@type` which contains the type URL. Example:
    
    package google.profile;
    message Person {
    string first_name = 1;
    string last_name = 2;
    }
    
    {
    "@type": "type.googleapis.com/google.profile.Person",
    "firstName": <string>,
    "lastName": <string>
    }
    
    If the embedded message type is well-known and has a custom JSON
    representation, that representation will be embedded adding a field
    `value` which holds the custom JSON in addition to the `@type`
    field. Example (for message [google.protobuf.Duration][]):
    
    {
    "@type": "type.googleapis.com/google.protobuf.Duration",
    "value": "1.212s"
    }
    """
    """A URL/resource name that uniquely identifies the type of the serialized
    protocol buffer message. This string must contain at least
    one "/" character. The last segment of the URL's path must represent
    the fully qualified name of the type (as in
    `path/google.protobuf.Duration`). The name should be in a canonical form
    (e.g., leading "." is not accepted).
    
    In practice, teams usually precompile into the binary all types that they
    expect it to use in the context of Any. However, for URLs which use the
    scheme `http`, `https`, or no scheme, one can optionally set up a type
    server that maps type URLs to message definitions as follows:
    
    * If no scheme is provided, `https` is assumed.
    * An HTTP GET on the URL must yield a [google.protobuf.Type][]
    value in binary format, or produce an error.
    * Applications are allowed to cache lookup results based on the
    URL, or have them precompiled into a binary to avoid any
    lookup. Therefore, binary compatibility needs to be preserved
    on changes to types. (Use versioned type names to manage
    breaking changes.)
    
    Note: this functionality is not currently available in the official
    protobuf release, and it is not used for type URLs beginning with
    type.googleapis.com.
    
    Schemes other than `http`, `https` (or the empty scheme) might be
    used with implementation specific semantics.
    """
    type_url: Optional[str]
    """Must be a valid serialized protocol buffer of the above specified type."""
    value: Optional[str]

    def __init__(self, type_url: Optional[str], value: Optional[str]) -> None:
        self.type_url = type_url
        self.value = value

    @staticmethod
    def from_dict(obj: Any) -> 'Detail':
        assert isinstance(obj, dict)
        type_url = from_union([from_str, from_none], obj.get("typeUrl"))
        value = from_union([from_str, from_none], obj.get("value"))
        return Detail(type_url, value)

    def to_dict(self) -> dict:
        result: dict = {}
        result["typeUrl"] = from_union([from_str, from_none], self.type_url)
        result["value"] = from_union([from_str, from_none], self.value)
        return result


class Status:
    """The status of the overall operation."""
    """The status code, which should be an enum value of [google.rpc.Code][google.rpc.Code]."""
    code: Optional[int]
    """A list of messages that carry the error details.  There is a common set of
    message types for APIs to use.
    """
    details: Optional[List[Detail]]
    """A developer-facing error message, which should be in English. Any
    user-facing error message should be localized and sent in the
    [google.rpc.Status.details][google.rpc.Status.details] field, or localized by the client.
    """
    message: Optional[str]

    def __init__(self, code: Optional[int], details: Optional[List[Detail]], message: Optional[str]) -> None:
        self.code = code
        self.details = details
        self.message = message

    @staticmethod
    def from_dict(obj: Any) -> 'Status':
        assert isinstance(obj, dict)
        code = from_union([from_int, from_none], obj.get("code"))
        details = from_union([lambda x: from_list(Detail.from_dict, x), from_none], obj.get("details"))
        message = from_union([from_str, from_none], obj.get("message"))
        return Status(code, details, message)

    def to_dict(self) -> dict:
        result: dict = {}
        result["code"] = from_union([from_int, from_none], self.code)
        result["details"] = from_union([lambda x: from_list(lambda x: to_class(Detail, x), x), from_none], self.details)
        result["message"] = from_union([from_str, from_none], self.message)
        return result


class ProtoPayload:
    """The log entry payload, which is always an AuditLog for Cloud Audit Log events."""
    """Authentication information."""
    authentication_info: Optional[AuthenticationInfo]
    """Authorization information. If there are multiple
    resources or permissions involved, then there is
    one AuthorizationInfo element for each {resource, permission} tuple.
    """
    authorization_info: Optional[List[AuthorizationInfo]]
    """Other service-specific data about the request, response, and other
    information associated with the current audited event.
    """
    metadata: Optional[Metadata]
    """The name of the service method or operation.
    For API calls, this should be the name of the API method.
    For example,
    
    "google.datastore.v1.Datastore.RunQuery"
    "google.logging.v1.LoggingService.DeleteLog"
    """
    method_name: Optional[str]
    """The number of items returned from a List or Query API method,
    if applicable.
    """
    num_response_items: Optional[str]
    """The operation request. This may not include all request parameters,
    such as those that are too large, privacy-sensitive, or duplicated
    elsewhere in the log record.
    It should never include user-generated data, such as file contents.
    When the JSON object represented here has a proto equivalent, the proto
    name will be indicated in the `@type` property.
    """
    request: Optional[Request]
    """Metadata about the operation."""
    request_metadata: Optional[RequestMetadata]
    """The resource location information."""
    resource_location: Optional[ResourceLocation]
    """The resource or collection that is the target of the operation.
    The name is a scheme-less URI, not including the API service name.
    For example:
    
    "shelves/SHELF_ID/books"
    "shelves/SHELF_ID/books/BOOK_ID"
    """
    resource_name: Optional[str]
    """The resource's original state before mutation. Present only for
    operations which have successfully modified the targeted resource(s).
    In general, this field should contain all changed fields, except those
    that are already been included in `request`, `response`, `metadata` or
    `service_data` fields.
    When the JSON object represented here has a proto equivalent,
    the proto name will be indicated in the `@type` property.
    """
    resource_original_state: Optional[ResourceOriginalState]
    """The operation response. This may not include all response elements,
    such as those that are too large, privacy-sensitive, or duplicated
    elsewhere in the log record.
    It should never include user-generated data, such as file contents.
    When the JSON object represented here has a proto equivalent, the proto
    name will be indicated in the `@type` property.
    """
    response: Optional[Response]
    """Deprecated, use `metadata` field instead.
    Other service-specific data about the request, response, and other
    activities.
    When the JSON object represented here has a proto equivalent, the proto
    name will be indicated in the `@type` property.
    """
    service_data: Optional[ServiceData]
    """The name of the API service performing the operation. For example,
    `"datastore.googleapis.com"`.
    """
    service_name: Optional[str]
    """The status of the overall operation."""
    status: Optional[Status]

    def __init__(self, authentication_info: Optional[AuthenticationInfo], authorization_info: Optional[List[AuthorizationInfo]], metadata: Optional[Metadata], method_name: Optional[str], num_response_items: Optional[str], request: Optional[Request], request_metadata: Optional[RequestMetadata], resource_location: Optional[ResourceLocation], resource_name: Optional[str], resource_original_state: Optional[ResourceOriginalState], response: Optional[Response], service_data: Optional[ServiceData], service_name: Optional[str], status: Optional[Status]) -> None:
        self.authentication_info = authentication_info
        self.authorization_info = authorization_info
        self.metadata = metadata
        self.method_name = method_name
        self.num_response_items = num_response_items
        self.request = request
        self.request_metadata = request_metadata
        self.resource_location = resource_location
        self.resource_name = resource_name
        self.resource_original_state = resource_original_state
        self.response = response
        self.service_data = service_data
        self.service_name = service_name
        self.status = status

    @staticmethod
    def from_dict(obj: Any) -> 'ProtoPayload':
        assert isinstance(obj, dict)
        authentication_info = from_union([AuthenticationInfo.from_dict, from_none], obj.get("authenticationInfo"))
        authorization_info = from_union([lambda x: from_list(AuthorizationInfo.from_dict, x), from_none], obj.get("authorizationInfo"))
        metadata = from_union([Metadata.from_dict, from_none], obj.get("metadata"))
        method_name = from_union([from_str, from_none], obj.get("methodName"))
        num_response_items = from_union([from_str, from_none], obj.get("numResponseItems"))
        request = from_union([Request.from_dict, from_none], obj.get("request"))
        request_metadata = from_union([RequestMetadata.from_dict, from_none], obj.get("requestMetadata"))
        resource_location = from_union([ResourceLocation.from_dict, from_none], obj.get("resourceLocation"))
        resource_name = from_union([from_str, from_none], obj.get("resourceName"))
        resource_original_state = from_union([ResourceOriginalState.from_dict, from_none], obj.get("resourceOriginalState"))
        response = from_union([Response.from_dict, from_none], obj.get("response"))
        service_data = from_union([ServiceData.from_dict, from_none], obj.get("serviceData"))
        service_name = from_union([from_str, from_none], obj.get("serviceName"))
        status = from_union([Status.from_dict, from_none], obj.get("status"))
        return ProtoPayload(authentication_info, authorization_info, metadata, method_name, num_response_items, request, request_metadata, resource_location, resource_name, resource_original_state, response, service_data, service_name, status)

    def to_dict(self) -> dict:
        result: dict = {}
        result["authenticationInfo"] = from_union([lambda x: to_class(AuthenticationInfo, x), from_none], self.authentication_info)
        result["authorizationInfo"] = from_union([lambda x: from_list(lambda x: to_class(AuthorizationInfo, x), x), from_none], self.authorization_info)
        result["metadata"] = from_union([lambda x: to_class(Metadata, x), from_none], self.metadata)
        result["methodName"] = from_union([from_str, from_none], self.method_name)
        result["numResponseItems"] = from_union([from_str, from_none], self.num_response_items)
        result["request"] = from_union([lambda x: to_class(Request, x), from_none], self.request)
        result["requestMetadata"] = from_union([lambda x: to_class(RequestMetadata, x), from_none], self.request_metadata)
        result["resourceLocation"] = from_union([lambda x: to_class(ResourceLocation, x), from_none], self.resource_location)
        result["resourceName"] = from_union([from_str, from_none], self.resource_name)
        result["resourceOriginalState"] = from_union([lambda x: to_class(ResourceOriginalState, x), from_none], self.resource_original_state)
        result["response"] = from_union([lambda x: to_class(Response, x), from_none], self.response)
        result["serviceData"] = from_union([lambda x: to_class(ServiceData, x), from_none], self.service_data)
        result["serviceName"] = from_union([from_str, from_none], self.service_name)
        result["status"] = from_union([lambda x: to_class(Status, x), from_none], self.status)
        return result


class Resource:
    """The monitored resource that produced this log entry.
    
    Example: a log entry that reports a database error would be associated with
    the monitored resource designating the particular database that reported
    the error.
    """
    """Values for all of the labels listed in the associated monitored
    resource descriptor. For example, Compute Engine VM instances use the
    labels `"project_id"`, `"instance_id"`, and `"zone"`.
    """
    labels: Optional[Dict[str, str]]
    """Required. The monitored resource type. For example, the type of a
    Compute Engine VM instance is `gce_instance`.
    """
    type: Optional[str]

    def __init__(self, labels: Optional[Dict[str, str]], type: Optional[str]) -> None:
        self.labels = labels
        self.type = type

    @staticmethod
    def from_dict(obj: Any) -> 'Resource':
        assert isinstance(obj, dict)
        labels = from_union([lambda x: from_dict(from_str, x), from_none], obj.get("labels"))
        type = from_union([from_str, from_none], obj.get("type"))
        return Resource(labels, type)

    def to_dict(self) -> dict:
        result: dict = {}
        result["labels"] = from_union([lambda x: from_dict(from_str, x), from_none], self.labels)
        result["type"] = from_union([from_str, from_none], self.type)
        return result


class LogEntryData:
    """The data within all Cloud Audit Logs log entry events."""
    """A unique identifier for the log entry."""
    insert_id: Optional[str]
    """A set of user-defined (key, value) data that provides additional
    information about the log entry.
    """
    labels: Optional[Dict[str, str]]
    """The resource name of the log to which this log entry belongs."""
    log_name: Optional[str]
    """Information about an operation associated with the log entry, if applicable."""
    operation: Optional[Operation]
    """The log entry payload, which is always an AuditLog for Cloud Audit Log events."""
    proto_payload: Optional[ProtoPayload]
    """The time the log entry was received by Logging."""
    receive_timestamp: Optional[datetime]
    """The monitored resource that produced this log entry.
    
    Example: a log entry that reports a database error would be associated with
    the monitored resource designating the particular database that reported
    the error.
    """
    resource: Optional[Resource]
    """The severity of the log entry."""
    severity: Union[int, None, str]
    """The span ID within the trace associated with the log entry, if any.
    
    For Trace spans, this is the same format that the Trace API v2 uses: a
    16-character hexadecimal encoding of an 8-byte array, such as
    `000000000000004a`.
    """
    span_id: Optional[str]
    """The time the event described by the log entry occurred."""
    timestamp: Optional[datetime]
    """Resource name of the trace associated with the log entry, if any. If it
    contains a relative resource name, the name is assumed to be relative to
    `//tracing.googleapis.com`. Example:
    `projects/my-projectid/traces/06796866738c859f2f19b7cfb3214824`
    """
    trace: Optional[str]

    def __init__(self, insert_id: Optional[str], labels: Optional[Dict[str, str]], log_name: Optional[str], operation: Optional[Operation], proto_payload: Optional[ProtoPayload], receive_timestamp: Optional[datetime], resource: Optional[Resource], severity: Union[int, None, str], span_id: Optional[str], timestamp: Optional[datetime], trace: Optional[str]) -> None:
        self.insert_id = insert_id
        self.labels = labels
        self.log_name = log_name
        self.operation = operation
        self.proto_payload = proto_payload
        self.receive_timestamp = receive_timestamp
        self.resource = resource
        self.severity = severity
        self.span_id = span_id
        self.timestamp = timestamp
        self.trace = trace

    @staticmethod
    def from_dict(obj: Any) -> 'LogEntryData':
        assert isinstance(obj, dict)
        insert_id = from_union([from_str, from_none], obj.get("insertId"))
        labels = from_union([lambda x: from_dict(from_str, x), from_none], obj.get("labels"))
        log_name = from_union([from_str, from_none], obj.get("logName"))
        operation = from_union([Operation.from_dict, from_none], obj.get("operation"))
        proto_payload = from_union([ProtoPayload.from_dict, from_none], obj.get("protoPayload"))
        receive_timestamp = from_union([from_datetime, from_none], obj.get("receiveTimestamp"))
        resource = from_union([Resource.from_dict, from_none], obj.get("resource"))
        severity = from_union([from_int, from_str, from_none], obj.get("severity"))
        span_id = from_union([from_str, from_none], obj.get("spanId"))
        timestamp = from_union([from_datetime, from_none], obj.get("timestamp"))
        trace = from_union([from_str, from_none], obj.get("trace"))
        return LogEntryData(insert_id, labels, log_name, operation, proto_payload, receive_timestamp, resource, severity, span_id, timestamp, trace)

    def to_dict(self) -> dict:
        result: dict = {}
        result["insertId"] = from_union([from_str, from_none], self.insert_id)
        result["labels"] = from_union([lambda x: from_dict(from_str, x), from_none], self.labels)
        result["logName"] = from_union([from_str, from_none], self.log_name)
        result["operation"] = from_union([lambda x: to_class(Operation, x), from_none], self.operation)
        result["protoPayload"] = from_union([lambda x: to_class(ProtoPayload, x), from_none], self.proto_payload)
        result["receiveTimestamp"] = from_union([lambda x: x.isoformat(), from_none], self.receive_timestamp)
        result["resource"] = from_union([lambda x: to_class(Resource, x), from_none], self.resource)
        result["severity"] = from_union([from_int, from_str, from_none], self.severity)
        result["spanId"] = from_union([from_str, from_none], self.span_id)
        result["timestamp"] = from_union([lambda x: x.isoformat(), from_none], self.timestamp)
        result["trace"] = from_union([from_str, from_none], self.trace)
        return result


def log_entry_data_from_dict(s: Any) -> LogEntryData:
    return LogEntryData.from_dict(s)


def log_entry_data_to_dict(x: LogEntryData) -> Any:
    return to_class(LogEntryData, x)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# To use this code, make sure you
#
#     import json
#
# and then, to convert JSON from a string, do
#
#     result = analytics_log_data_from_dict(json.loads(json_string))

from typing import Optional, Any, Dict, List, TypeVar, Callable, Type, cast


T = TypeVar("T")


def from_float(x: Any) -> float:
    assert isinstance(x, (float, int)) and not isinstance(x, bool)
    return float(x)


def from_none(x: Any) -> Any:
    assert x is None
    return x


def from_union(fs, x):
    for f in fs:
        try:
            return f(x)
        except:
            pass
    assert False


def from_str(x: Any) -> str:
    assert isinstance(x, str)
    return x


def to_float(x: Any) -> float:
    assert isinstance(x, float)
    return x


def from_dict(f: Callable[[Any], T], x: Any) -> Dict[str, T]:
    assert isinstance(x, dict)
    return { k: f(v) for (k, v) in x.items() }


def to_class(c: Type[T], x: Any) -> dict:
    assert isinstance(x, c)
    return cast(Any, x).to_dict()


def from_int(x: Any) -> int:
    assert isinstance(x, int) and not isinstance(x, bool)
    return x


def from_bool(x: Any) -> bool:
    assert isinstance(x, bool)
    return x


def from_list(f: Callable[[Any], T], x: Any) -> List[T]:
    assert isinstance(x, list)
    return [f(y) for y in x]


class AnalyticsValue:
    """Value for Event Params and UserProperty can be of type string or int or
    float or double.
    """
    double_value: Optional[float]
    float_value: Optional[float]
    int_value: Optional[str]
    string_value: Optional[str]

    def __init__(self, double_value: Optional[float], float_value: Optional[float], int_value: Optional[str], string_value: Optional[str]) -> None:
        self.double_value = double_value
        self.float_value = float_value
        self.int_value = int_value
        self.string_value = string_value

    @staticmethod
    def from_dict(obj: Any) -> 'AnalyticsValue':
        assert isinstance(obj, dict)
        double_value = from_union([from_float, from_none], obj.get("doubleValue"))
        float_value = from_union([from_float, from_none], obj.get("floatValue"))
        int_value = from_union([from_str, from_none], obj.get("intValue"))
        string_value = from_union([from_str, from_none], obj.get("stringValue"))
        return AnalyticsValue(double_value, float_value, int_value, string_value)

    def to_dict(self) -> dict:
        result: dict = {}
        result["doubleValue"] = from_union([to_float, from_none], self.double_value)
        result["floatValue"] = from_union([to_float, from_none], self.float_value)
        result["intValue"] = from_union([from_str, from_none], self.int_value)
        result["stringValue"] = from_union([from_str, from_none], self.string_value)
        return result


class EventDim:
    """Message containing information pertaining to the event."""
    """The date on which this event was logged.
    (YYYYMMDD format in the registered timezone of your app.)
    """
    date: Optional[str]
    """The name of this event."""
    name: Optional[str]
    """A repeated record of the parameters associated with this event."""
    params: Optional[Dict[str, AnalyticsValue]]
    """UTC client time when the previous event happened."""
    previous_timestamp_micros: Optional[str]
    """UTC client time when the event happened."""
    timestamp_micros: Optional[str]
    """Value param in USD."""
    value_in_usd: Optional[float]

    def __init__(self, date: Optional[str], name: Optional[str], params: Optional[Dict[str, AnalyticsValue]], previous_timestamp_micros: Optional[str], timestamp_micros: Optional[str], value_in_usd: Optional[float]) -> None:
        self.date = date
        self.name = name
        self.params = params
        self.previous_timestamp_micros = previous_timestamp_micros
        self.timestamp_micros = timestamp_micros
        self.value_in_usd = value_in_usd

    @staticmethod
    def from_dict(obj: Any) -> 'EventDim':
        assert isinstance(obj, dict)
        date = from_union([from_str, from_none], obj.get("date"))
        name = from_union([from_str, from_none], obj.get("name"))
        params = from_union([lambda x: from_dict(AnalyticsValue.from_dict, x), from_none], obj.get("params"))
        previous_timestamp_micros = from_union([from_str, from_none], obj.get("previousTimestampMicros"))
        timestamp_micros = from_union([from_str, from_none], obj.get("timestampMicros"))
        value_in_usd = from_union([from_float, from_none], obj.get("valueInUsd"))
        return EventDim(date, name, params, previous_timestamp_micros, timestamp_micros, value_in_usd)

    def to_dict(self) -> dict:
        result: dict = {}
        result["date"] = from_union([from_str, from_none], self.date)
        result["name"] = from_union([from_str, from_none], self.name)
        result["params"] = from_union([lambda x: from_dict(lambda x: to_class(AnalyticsValue, x), x), from_none], self.params)
        result["previousTimestampMicros"] = from_union([from_str, from_none], self.previous_timestamp_micros)
        result["timestampMicros"] = from_union([from_str, from_none], self.timestamp_micros)
        result["valueInUsd"] = from_union([to_float, from_none], self.value_in_usd)
        return result


class AppInfo:
    """App information."""
    """Unique application identifier within an app store."""
    app_id: Optional[str]
    """Unique id for this instance of the app.
    Example: "71683BF9FA3B4B0D9535A1F05188BAF3"
    """
    app_instance_id: Optional[str]
    """The app platform.
    Eg "ANDROID", "IOS".
    """
    app_platform: Optional[str]
    """The identifier of the store that installed the app.
    Eg. "com.sec.android.app.samsungapps", "com.amazon.venezia",
    "com.nokia.nstore"
    """
    app_store: Optional[str]
    """The app's version name
    Examples: "1.0", "4.3.1.1.213361", "2.3 (1824253)", "v1.8b22p6"
    """
    app_version: Optional[str]

    def __init__(self, app_id: Optional[str], app_instance_id: Optional[str], app_platform: Optional[str], app_store: Optional[str], app_version: Optional[str]) -> None:
        self.app_id = app_id
        self.app_instance_id = app_instance_id
        self.app_platform = app_platform
        self.app_store = app_store
        self.app_version = app_version

    @staticmethod
    def from_dict(obj: Any) -> 'AppInfo':
        assert isinstance(obj, dict)
        app_id = from_union([from_str, from_none], obj.get("appId"))
        app_instance_id = from_union([from_str, from_none], obj.get("appInstanceId"))
        app_platform = from_union([from_str, from_none], obj.get("appPlatform"))
        app_store = from_union([from_str, from_none], obj.get("appStore"))
        app_version = from_union([from_str, from_none], obj.get("appVersion"))
        return AppInfo(app_id, app_instance_id, app_platform, app_store, app_version)

    def to_dict(self) -> dict:
        result: dict = {}
        result["appId"] = from_union([from_str, from_none], self.app_id)
        result["appInstanceId"] = from_union([from_str, from_none], self.app_instance_id)
        result["appPlatform"] = from_union([from_str, from_none], self.app_platform)
        result["appStore"] = from_union([from_str, from_none], self.app_store)
        result["appVersion"] = from_union([from_str, from_none], self.app_version)
        return result


class BundleInfo:
    """Information regarding the bundle in which these events were uploaded."""
    """Monotonically increasing index for each bundle set by SDK."""
    bundle_sequence_id: Optional[int]
    """Timestamp offset between collection time and upload time."""
    server_timestamp_offset_micros: Optional[str]

    def __init__(self, bundle_sequence_id: Optional[int], server_timestamp_offset_micros: Optional[str]) -> None:
        self.bundle_sequence_id = bundle_sequence_id
        self.server_timestamp_offset_micros = server_timestamp_offset_micros

    @staticmethod
    def from_dict(obj: Any) -> 'BundleInfo':
        assert isinstance(obj, dict)
        bundle_sequence_id = from_union([from_int, from_none], obj.get("bundleSequenceId"))
        server_timestamp_offset_micros = from_union([from_str, from_none], obj.get("serverTimestampOffsetMicros"))
        return BundleInfo(bundle_sequence_id, server_timestamp_offset_micros)

    def to_dict(self) -> dict:
        result: dict = {}
        result["bundleSequenceId"] = from_union([from_int, from_none], self.bundle_sequence_id)
        result["serverTimestampOffsetMicros"] = from_union([from_str, from_none], self.server_timestamp_offset_micros)
        return result


class DeviceInfo:
    """Device information."""
    """Device category.
    Eg. tablet or mobile.
    """
    device_category: Optional[str]
    """Vendor specific device identifier. This is IDFV on iOS. Not used for
    Android.
    Example: "599F9C00-92DC-4B5C-9464-7971F01F8370"
    """
    device_id: Optional[str]
    """Device model.
    Eg. GT-I9192
    """
    device_model: Optional[str]
    """The timezone of the device when data was uploaded as seconds skew from UTC."""
    device_time_zone_offset_seconds: Optional[int]
    """The device's Limit Ad Tracking setting.
    When true, we cannot use device_id for remarketing, demographics or
    influencing ads serving behaviour. However, we can use device_id for
    conversion tracking and campaign attribution.
    """
    limited_ad_tracking: Optional[bool]
    """Device brand name.
    Eg. Samsung, HTC, etc.
    """
    mobile_brand_name: Optional[str]
    """Device marketing name.
    Eg. Galaxy S4 Mini
    """
    mobile_marketing_name: Optional[str]
    """Device model name.
    Eg. GT-I9192
    """
    mobile_model_name: Optional[str]
    """Device OS version when data capture ended.
    Eg. 4.4.2
    """
    platform_version: Optional[str]
    """The type of the resettable_device_id is always IDFA on iOS and AdId
    on Android.
    Example: "71683BF9-FA3B-4B0D-9535-A1F05188BAF3"
    """
    resettable_device_id: Optional[str]
    """The user language.
    Eg. "en-us", "en-za", "zh-tw", "jp"
    """
    user_default_language: Optional[str]

    def __init__(self, device_category: Optional[str], device_id: Optional[str], device_model: Optional[str], device_time_zone_offset_seconds: Optional[int], limited_ad_tracking: Optional[bool], mobile_brand_name: Optional[str], mobile_marketing_name: Optional[str], mobile_model_name: Optional[str], platform_version: Optional[str], resettable_device_id: Optional[str], user_default_language: Optional[str]) -> None:
        self.device_category = device_category
        self.device_id = device_id
        self.device_model = device_model
        self.device_time_zone_offset_seconds = device_time_zone_offset_seconds
        self.limited_ad_tracking = limited_ad_tracking
        self.mobile_brand_name = mobile_brand_name
        self.mobile_marketing_name = mobile_marketing_name
        self.mobile_model_name = mobile_model_name
        self.platform_version = platform_version
        self.resettable_device_id = resettable_device_id
        self.user_default_language = user_default_language

    @staticmethod
    def from_dict(obj: Any) -> 'DeviceInfo':
        assert isinstance(obj, dict)
        device_category = from_union([from_str, from_none], obj.get("deviceCategory"))
        device_id = from_union([from_str, from_none], obj.get("deviceId"))
        device_model = from_union([from_str, from_none], obj.get("deviceModel"))
        device_time_zone_offset_seconds = from_union([from_int, from_none], obj.get("deviceTimeZoneOffsetSeconds"))
        limited_ad_tracking = from_union([from_bool, from_none], obj.get("limitedAdTracking"))
        mobile_brand_name = from_union([from_str, from_none], obj.get("mobileBrandName"))
        mobile_marketing_name = from_union([from_str, from_none], obj.get("mobileMarketingName"))
        mobile_model_name = from_union([from_str, from_none], obj.get("mobileModelName"))
        platform_version = from_union([from_str, from_none], obj.get("platformVersion"))
        resettable_device_id = from_union([from_str, from_none], obj.get("resettableDeviceId"))
        user_default_language = from_union([from_str, from_none], obj.get("userDefaultLanguage"))
        return DeviceInfo(device_category, device_id, device_model, device_time_zone_offset_seconds, limited_ad_tracking, mobile_brand_name, mobile_marketing_name, mobile_model_name, platform_version, resettable_device_id, user_default_language)

    def to_dict(self) -> dict:
        result: dict = {}
        result["deviceCategory"] = from_union([from_str, from_none], self.device_category)
        result["deviceId"] = from_union([from_str, from_none], self.device_id)
        result["deviceModel"] = from_union([from_str, from_none], self.device_model)
        result["deviceTimeZoneOffsetSeconds"] = from_union([from_int, from_none], self.device_time_zone_offset_seconds)
        result["limitedAdTracking"] = from_union([from_bool, from_none], self.limited_ad_tracking)
        result["mobileBrandName"] = from_union([from_str, from_none], self.mobile_brand_name)
        result["mobileMarketingName"] = from_union([from_str, from_none], self.mobile_marketing_name)
        result["mobileModelName"] = from_union([from_str, from_none], self.mobile_model_name)
        result["platformVersion"] = from_union([from_str, from_none], self.platform_version)
        result["resettableDeviceId"] = from_union([from_str, from_none], self.resettable_device_id)
        result["userDefaultLanguage"] = from_union([from_str, from_none], self.user_default_language)
        return result


class GeoInfo:
    """User's geographic information."""
    """The geographic city.
    Eg. Sao Paulo
    """
    city: Optional[str]
    """The geographic continent.
    Eg. Americas
    """
    continent: Optional[str]
    """The geographic country.
    Eg. Brazil
    """
    country: Optional[str]
    """The geographic region.
    Eg. State of Sao Paulo
    """
    region: Optional[str]

    def __init__(self, city: Optional[str], continent: Optional[str], country: Optional[str], region: Optional[str]) -> None:
        self.city = city
        self.continent = continent
        self.country = country
        self.region = region

    @staticmethod
    def from_dict(obj: Any) -> 'GeoInfo':
        assert isinstance(obj, dict)
        city = from_union([from_str, from_none], obj.get("city"))
        continent = from_union([from_str, from_none], obj.get("continent"))
        country = from_union([from_str, from_none], obj.get("country"))
        region = from_union([from_str, from_none], obj.get("region"))
        return GeoInfo(city, continent, country, region)

    def to_dict(self) -> dict:
        result: dict = {}
        result["city"] = from_union([from_str, from_none], self.city)
        result["continent"] = from_union([from_str, from_none], self.continent)
        result["country"] = from_union([from_str, from_none], self.country)
        result["region"] = from_union([from_str, from_none], self.region)
        return result


class LtvInfo:
    """Lifetime Value information about this user."""
    """The currency corresponding to the revenue."""
    currency: Optional[str]
    """The Lifetime Value revenue of this user."""
    revenue: Optional[float]

    def __init__(self, currency: Optional[str], revenue: Optional[float]) -> None:
        self.currency = currency
        self.revenue = revenue

    @staticmethod
    def from_dict(obj: Any) -> 'LtvInfo':
        assert isinstance(obj, dict)
        currency = from_union([from_str, from_none], obj.get("currency"))
        revenue = from_union([from_float, from_none], obj.get("revenue"))
        return LtvInfo(currency, revenue)

    def to_dict(self) -> dict:
        result: dict = {}
        result["currency"] = from_union([from_str, from_none], self.currency)
        result["revenue"] = from_union([to_float, from_none], self.revenue)
        return result


class TrafficSource:
    """Information about marketing campaign which acquired the user."""
    """The name of the campaign which acquired the user."""
    user_acquired_campaign: Optional[str]
    """The name of the medium which acquired the user."""
    user_acquired_medium: Optional[str]
    """The name of the network which acquired the user."""
    user_acquired_source: Optional[str]

    def __init__(self, user_acquired_campaign: Optional[str], user_acquired_medium: Optional[str], user_acquired_source: Optional[str]) -> None:
        self.user_acquired_campaign = user_acquired_campaign
        self.user_acquired_medium = user_acquired_medium
        self.user_acquired_source = user_acquired_source

    @staticmethod
    def from_dict(obj: Any) -> 'TrafficSource':
        assert isinstance(obj, dict)
        user_acquired_campaign = from_union([from_str, from_none], obj.get("userAcquiredCampaign"))
        user_acquired_medium = from_union([from_str, from_none], obj.get("userAcquiredMedium"))
        user_acquired_source = from_union([from_str, from_none], obj.get("userAcquiredSource"))
        return TrafficSource(user_acquired_campaign, user_acquired_medium, user_acquired_source)

    def to_dict(self) -> dict:
        result: dict = {}
        result["userAcquiredCampaign"] = from_union([from_str, from_none], self.user_acquired_campaign)
        result["userAcquiredMedium"] = from_union([from_str, from_none], self.user_acquired_medium)
        result["userAcquiredSource"] = from_union([from_str, from_none], self.user_acquired_source)
        return result


class Value:
    """Last set value of user property.
    
    Value for Event Params and UserProperty can be of type string or int or
    float or double.
    """
    double_value: Optional[float]
    float_value: Optional[float]
    int_value: Optional[str]
    string_value: Optional[str]

    def __init__(self, double_value: Optional[float], float_value: Optional[float], int_value: Optional[str], string_value: Optional[str]) -> None:
        self.double_value = double_value
        self.float_value = float_value
        self.int_value = int_value
        self.string_value = string_value

    @staticmethod
    def from_dict(obj: Any) -> 'Value':
        assert isinstance(obj, dict)
        double_value = from_union([from_float, from_none], obj.get("doubleValue"))
        float_value = from_union([from_float, from_none], obj.get("floatValue"))
        int_value = from_union([from_str, from_none], obj.get("intValue"))
        string_value = from_union([from_str, from_none], obj.get("stringValue"))
        return Value(double_value, float_value, int_value, string_value)

    def to_dict(self) -> dict:
        result: dict = {}
        result["doubleValue"] = from_union([to_float, from_none], self.double_value)
        result["floatValue"] = from_union([to_float, from_none], self.float_value)
        result["intValue"] = from_union([from_str, from_none], self.int_value)
        result["stringValue"] = from_union([from_str, from_none], self.string_value)
        return result


class UserProperty:
    """Index for user property (one-based)."""
    index: Optional[int]
    """UTC client time when user property was last set."""
    set_timestamp_usec: Optional[str]
    """Last set value of user property."""
    value: Optional[Value]

    def __init__(self, index: Optional[int], set_timestamp_usec: Optional[str], value: Optional[Value]) -> None:
        self.index = index
        self.set_timestamp_usec = set_timestamp_usec
        self.value = value

    @staticmethod
    def from_dict(obj: Any) -> 'UserProperty':
        assert isinstance(obj, dict)
        index = from_union([from_int, from_none], obj.get("index"))
        set_timestamp_usec = from_union([from_str, from_none], obj.get("setTimestampUsec"))
        value = from_union([Value.from_dict, from_none], obj.get("value"))
        return UserProperty(index, set_timestamp_usec, value)

    def to_dict(self) -> dict:
        result: dict = {}
        result["index"] = from_union([from_int, from_none], self.index)
        result["setTimestampUsec"] = from_union([from_str, from_none], self.set_timestamp_usec)
        result["value"] = from_union([lambda x: to_class(Value, x), from_none], self.value)
        return result


class UserDim:
    """User related dimensions."""
    """App information."""
    app_info: Optional[AppInfo]
    """Information regarding the bundle in which these events were uploaded."""
    bundle_info: Optional[BundleInfo]
    """Device information."""
    device_info: Optional[DeviceInfo]
    """The time (in microseconds) at which the user first opened the app."""
    first_open_timestamp_micros: Optional[str]
    """User's geographic information."""
    geo_info: Optional[GeoInfo]
    """Lifetime Value information about this user."""
    ltv_info: Optional[LtvInfo]
    """Information about marketing campaign which acquired the user."""
    traffic_source: Optional[TrafficSource]
    """The user ID set via the setUserId API."""
    user_id: Optional[str]
    """A repeated record of user properties set with the setUserProperty API.
    https://firebase.google.com/docs/analytics/android/properties
    """
    user_properties: Optional[Dict[str, UserProperty]]

    def __init__(self, app_info: Optional[AppInfo], bundle_info: Optional[BundleInfo], device_info: Optional[DeviceInfo], first_open_timestamp_micros: Optional[str], geo_info: Optional[GeoInfo], ltv_info: Optional[LtvInfo], traffic_source: Optional[TrafficSource], user_id: Optional[str], user_properties: Optional[Dict[str, UserProperty]]) -> None:
        self.app_info = app_info
        self.bundle_info = bundle_info
        self.device_info = device_info
        self.first_open_timestamp_micros = first_open_timestamp_micros
        self.geo_info = geo_info
        self.ltv_info = ltv_info
        self.traffic_source = traffic_source
        self.user_id = user_id
        self.user_properties = user_properties

    @staticmethod
    def from_dict(obj: Any) -> 'UserDim':
        assert isinstance(obj, dict)
        app_info = from_union([AppInfo.from_dict, from_none], obj.get("appInfo"))
        bundle_info = from_union([BundleInfo.from_dict, from_none], obj.get("bundleInfo"))
        device_info = from_union([DeviceInfo.from_dict, from_none], obj.get("deviceInfo"))
        first_open_timestamp_micros = from_union([from_str, from_none], obj.get("firstOpenTimestampMicros"))
        geo_info = from_union([GeoInfo.from_dict, from_none], obj.get("geoInfo"))
        ltv_info = from_union([LtvInfo.from_dict, from_none], obj.get("ltvInfo"))
        traffic_source = from_union([TrafficSource.from_dict, from_none], obj.get("trafficSource"))
        user_id = from_union([from_str, from_none], obj.get("userId"))
        user_properties = from_union([lambda x: from_dict(UserProperty.from_dict, x), from_none], obj.get("userProperties"))
        return UserDim(app_info, bundle_info, device_info, first_open_timestamp_micros, geo_info, ltv_info, traffic_source, user_id, user_properties)

    def to_dict(self) -> dict:
        result: dict = {}
        result["appInfo"] = from_union([lambda x: to_class(AppInfo, x), from_none], self.app_info)
        result["bundleInfo"] = from_union([lambda x: to_class(BundleInfo, x), from_none], self.bundle_info)
        result["deviceInfo"] = from_union([lambda x: to_class(DeviceInfo, x), from_none], self.device_info)
        result["firstOpenTimestampMicros"] = from_union([from_str, from_none], self.first_open_timestamp_micros)
        result["geoInfo"] = from_union([lambda x: to_class(GeoInfo, x), from_none], self.geo_info)
        result["ltvInfo"] = from_union([lambda x: to_class(LtvInfo, x), from_none], self.ltv_info)
        result["trafficSource"] = from_union([lambda x: to_class(TrafficSource, x), from_none], self.traffic_source)
        result["userId"] = from_union([from_str, from_none], self.user_id)
        result["userProperties"] = from_union([lambda x: from_dict(lambda x: to_class(UserProperty, x), x), from_none], self.user_properties)
        return result


class AnalyticsLogData:
    """The data within Firebase Analytics log events."""
    """A repeated record of event related dimensions."""
    event_dim: Optional[List[EventDim]]
    """User related dimensions."""
    user_dim: Optional[UserDim]

    def __init__(self, event_dim: Optional[List[EventDim]], user_dim: Optional[UserDim]) -> None:
        self.event_dim = event_dim
        self.user_dim = user_dim

    @staticmethod
    def from_dict(obj: Any) -> 'AnalyticsLogData':
        assert isinstance(obj, dict)
        event_dim = from_union([lambda x: from_list(EventDim.from_dict, x), from_none], obj.get("eventDim"))
        user_dim = from_union([UserDim.from_dict, from_none], obj.get("userDim"))
        return AnalyticsLogData(event_dim, user_dim)

    def to_dict(self) -> dict:
        result: dict = {}
        result["eventDim"] = from_union([lambda x: from_list(lambda x: to_class(EventDim, x), x), from_none], self.event_dim)
        result["userDim"] = from_union([lambda x: to_class(UserDim, x), from_none], self.user_dim)
        return result


def analytics_log_data_from_dict(s: Any) -> AnalyticsLogData:
    return AnalyticsLogData.from_dict(s)


def analytics_log_data_to_dict(x: AnalyticsLogData) -> Any:
    return to_class(AnalyticsLogData, x)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# To use this code, make sure you
#
#     import json
#
# and then, to convert JSON from a string, do
#
#     result = scheduler_job_data_from_dict(json.loads(json_string))

from typing import Optional, Any, TypeVar, Type, cast


T = TypeVar("T")


def from_str(x: Any) -> str:
    assert isinstance(x, str)
    return x


def from_none(x: Any) -> Any:
    assert x is None
    return x


def from_union(fs, x):
    for f in fs:
        try:
            return f(x)
        except:
            pass
    assert False


def to_class(c: Type[T], x: Any) -> dict:
    assert isinstance(x, c)
    return cast(Any, x).to_dict()


class SchedulerJobData:
    """Scheduler job data."""
    """The custom data the user specified when creating the scheduler source."""
    custom_data: Optional[str]

    def __init__(self, custom_data: Optional[str]) -> None:
        self.custom_data = custom_data

    @staticmethod
    def from_dict(obj: Any) -> 'SchedulerJobData':
        assert isinstance(obj, dict)
        custom_data = from_union([from_str, from_none], obj.get("customData"))
        return SchedulerJobData(custom_data)

    def to_dict(self) -> dict:
        result: dict = {}
        result["customData"] = from_union([from_str, from_none], self.custom_data)
        return result


def scheduler_job_data_from_dict(s: Any) -> SchedulerJobData:
    return SchedulerJobData.from_dict(s)


def scheduler_job_data_to_dict(x: SchedulerJobData) -> Any:
    return to_class(SchedulerJobData, x)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# To use this code, make sure you
#
#     import json
#
# and then, to convert JSON from a string, do
#
#     result = reference_event_data_from_dict(json.loads(json_string))

from typing import Optional, Dict, Any, TypeVar, Callable, Type, cast


T = TypeVar("T")


def from_none(x: Any) -> Any:
    assert x is None
    return x


def from_dict(f: Callable[[Any], T], x: Any) -> Dict[str, T]:
    assert isinstance(x, dict)
    return { k: f(v) for (k, v) in x.items() }


def from_union(fs, x):
    for f in fs:
        try:
            return f(x)
        except:
            pass
    assert False


def to_class(c: Type[T], x: Any) -> dict:
    assert isinstance(x, c)
    return cast(Any, x).to_dict()


class ReferenceEventData:
    """The data within all Firebase Real Time Database reference events."""
    data: Optional[Dict[str, Any]]
    delta: Optional[Dict[str, Any]]

    def __init__(self, data: Optional[Dict[str, Any]], delta: Optional[Dict[str, Any]]) -> None:
        self.data = data
        self.delta = delta

    @staticmethod
    def from_dict(obj: Any) -> 'ReferenceEventData':
        assert isinstance(obj, dict)
        data = from_union([from_none, lambda x: from_dict(lambda x: x, x)], obj.get("data"))
        delta = from_union([from_none, lambda x: from_dict(lambda x: x, x)], obj.get("delta"))
        return ReferenceEventData(data, delta)

    def to_dict(self) -> dict:
        result: dict = {}
        result["data"] = from_union([from_none, lambda x: from_dict(lambda x: x, x)], self.data)
        result["delta"] = from_union([from_none, lambda x: from_dict(lambda x: x, x)], self.delta)
        return result


def reference_event_data_from_dict(s: Any) -> ReferenceEventData:
    return ReferenceEventData.from_dict(s)


def reference_event_data_to_dict(x: ReferenceEventData) -> Any:
    return to_class(ReferenceEventData, x)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# This code parses date/times, so please
#
#     pip install python-dateutil
#
# To use this code, make sure you
#
#     import json
#
# and then, to convert JSON from a string, do
#
#     result = remote_config_event_data_from_dict(json.loads(json_string))

from typing import Optional, Any, Union, TypeVar, Type, cast
from datetime import datetime
import dateutil.parser


T = TypeVar("T")


def from_str(x: Any) -> str:
    assert isinstance(x, str)
    return x


def from_none(x: Any) -> Any:
    assert x is None
    return x


def from_union(fs, x):
    for f in fs:
        try:
            return f(x)
        except:
            pass
    assert False


def from_int(x: Any) -> int:
    assert isinstance(x, int) and not isinstance(x, bool)
    return x


def from_datetime(x: Any) -> datetime:
    return dateutil.parser.parse(x)


def to_class(c: Type[T], x: Any) -> dict:
    assert isinstance(x, c)
    return cast(Any, x).to_dict()


class UpdateUser:
    """Aggregation of all metadata fields about the account that performed the update."""
    """Email address."""
    email: Optional[str]
    """Image URL."""
    image_url: Optional[str]
    """Display name."""
    name: Optional[str]

    def __init__(self, email: Optional[str], image_url: Optional[str], name: Optional[str]) -> None:
        self.email = email
        self.image_url = image_url
        self.name = name

    @staticmethod
    def from_dict(obj: Any) -> 'UpdateUser':
        assert isinstance(obj, dict)
        email = from_union([from_str, from_none], obj.get("email"))
        image_url = from_union([from_str, from_none], obj.get("imageUrl"))
        name = from_union([from_str, from_none], obj.get("name"))
        return UpdateUser(email, image_url, name)

    def to_dict(self) -> dict:
        result: dict = {}
        result["email"] = from_union([from_str, from_none], self.email)
        result["imageUrl"] = from_union([from_str, from_none], self.image_url)
        result["name"] = from_union([from_str, from_none], self.name)
        return result


class RemoteConfigEventData:
    """The data within all Firebase Remote Config events."""
    """The user-provided description of the corresponding Remote Config template."""
    description: Optional[str]
    """Only present if this version is the result of a rollback, and will be the
    version number of the Remote Config template that was rolled-back to.
    """
    rollback_source: Optional[str]
    """Where the update action originated."""
    update_origin: Union[int, None, str]
    """When the Remote Config template was written to the Remote Config server."""
    update_time: Optional[datetime]
    """What type of update was made."""
    update_type: Union[int, None, str]
    """Aggregation of all metadata fields about the account that performed the update."""
    update_user: Optional[UpdateUser]
    """The version number of the version's corresponding Remote Config template."""
    version_number: Optional[str]

    def __init__(self, description: Optional[str], rollback_source: Optional[str], update_origin: Union[int, None, str], update_time: Optional[datetime], update_type: Union[int, None, str], update_user: Optional[UpdateUser], version_number: Optional[str]) -> None:
        self.description = description
        self.rollback_source = rollback_source
        self.update_origin = update_origin
        self.update_time = update_time
        self.update_type = update_type
        self.update_user = update_user
        self.version_number = version_number

    @staticmethod
    def from_dict(obj: Any) -> 'RemoteConfigEventData':
        assert isinstance(obj, dict)
        description = from_union([from_str, from_none], obj.get("description"))
        rollback_source = from_union([from_str, from_none], obj.get("rollbackSource"))
        update_origin = from_union([from_int, from_str, from_none], obj.get("updateOrigin"))
        update_time = from_union([from_datetime, from_none], obj.get("updateTime"))
        update_type = from_union([from_int, from_str, from_none], obj.get("updateType"))
        update_user = from_union([UpdateUser.from_dict, from_none], obj.get("updateUser"))
        version_number = from_union([from_str, from_none], obj.get("versionNumber"))
        return RemoteConfigEventData(description, rollback_source, update_origin, update_time, update_type, update_user, version_number)

    def to_dict(self) -> dict:
        result: dict = {}
        result["description"] = from_union([from_str, from_none], self.description)
        result["rollbackSource"] = from_union([from_str, from_none], self.rollback_source)
        result["updateOrigin"] = from_union([from_int, from_str, from_none], self.update_origin)
        result["updateTime"] = from_union([lambda x: x.isoformat(), from_none], self.update_time)
        result["updateType"] = from_union([from_int, from_str, from_none], self.update_type)
        result["updateUser"] = from_union([lambda x: to_class(UpdateUser, x), from_none], self.update_user)
        result["versionNumber"] = from_union([from_str, from_none], self.version_number)
        return result


def remote_config_event_data_from_dict(s: Any) -> RemoteConfigEventData:
    return RemoteConfigEventData.from_dict(s)


def remote_config_event_data_to_dict(x: RemoteConfigEventData) -> Any:
    return to_class(RemoteConfigEventData, x)



import os
from setuptools import find_packages, setup

release_status = "Development Status :: 3 - Alpha"

package_root = os.path.abspath(os.path.dirname(__file__))
readme_filename = os.path.join(package_root, "README.md")
with open(readme_filename, encoding="utf-8") as readme_file:
    readme = readme_file.read()

setup(
    name="google-events",
    version = "0.1.2",
    description="A collection of first party Google Cloud Platform event objects.",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/googleapis/google-cloudevents",
    project_urls={
        "Source": "https://github.com/googleapis/google-cloudevents-python",
        "Documentation": "https://github.com/googleapis/google-cloudevents-python",
        "Issue Tracker": "https://github.com/googleapis/google-cloudevents-python/issues",
    },
    license="Apache License 2.0",
    author="Google LLC",
    author_email="googleapis-packages@google.com",
    classifiers=[
        release_status,
        "Environment :: Web Environment",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Topic :: Internet",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    packages=find_packages("src"),
    package_dir={"": "src"},
    python_requires=">=3.7",
    install_requires=['python-dateutil==2.8.1'],
    extras_require={"dev": ["pylint", "pytest", "black", "stringcase"]}
)
